# refer from,
# https://github.com/haeusser/learning_by_association/blob/master/semisup/backend.py

import numpy as np
import tensorflow as tf

def loss (a, b, labels, visit_weight=1.0, walker_weight=1.0):
    """Add semi-supervised classification loss to the model.

    The loss constist of two terms: "walker" and "visit".

    Args:
      a: [N, emb_size] tensor with supervised embedding vectors.
      b: [M, emb_size] tensor with unsupervised embedding vectors.
      labels : [N] tensor with labels for supervised embeddings.
      walker_weight: Weight coefficient of the "walker" loss.
      visit_weight: Weight coefficient of the "visit" loss.
    """

    equality_matrix = tf.equal (tf.reshape(labels, [-1, 1]), labels)
    equality_matrix = tf.cast (equality_matrix, tf.float32)
    p_target = equality_matrix / tf.reduce_sum (equality_matrix, [1], keepdims=True)
    
    match_ab = tf.matmul(a, b, transpose_b=True, name='match_ab')
    p_ab = tf.nn.softmax(match_ab, name='p_ab')
    p_ba = tf.nn.softmax(tf.transpose(match_ab), name='p_ba')
    p_aba = tf.matmul(p_ab, p_ba, name='p_aba')
   
    loss_walker = tf.reduce_mean (tf.nn.softmax_cross_entropy_with_logits_v2 (
        labels = p_target,
        logits = tf.log (1e-8 + p_aba)        
    )) * walker_weight
    
    return loss_walker, visit_loss (p_ab, visit_weight), estimate_error (p_aba, equality_matrix)

def visit_loss (p, weight=1.0):
    """Add the "visit" loss to the model.

    Args:
      p: [N, M] tensor. Each row must be a valid probability distribution
          (i.e. sum to 1.0)
      weight: Loss weight.
    """
    visit_probability = tf.reduce_mean(p, [0], keepdims=True, name='visit_prob')
    t_nb = tf.shape(p)[1]

    return tf.reduce_mean (tf.nn.softmax_cross_entropy_with_logits_v2 (
        labels = tf.fill([1, t_nb], 1.0 / tf.cast(t_nb, tf.float32)),
        logits = tf.log(1e-8 + visit_probability)
    )) * weight
    
    
def estimate_error (p_aba, equality_matrix):
    """Adds "walker" loss statistics to the graph.
    Args:
      p_aba: [N, N] matrix, where element [i, j] corresponds to the
          probalility of the round-trip between supervised samples i and j.
          Sum of each row of 'p_aba' must be equal to one.
      equality_matrix: [N, N] boolean matrix, [i,j] is True, when samples
          i and j belong to the same class.
    """
    # Using the square root of the correct round trip probalilty as an estimate
    # of the current classifier accuracy.
    per_row_accuracy = 1.0 - tf.reduce_sum((equality_matrix * p_aba), 1)**0.5
    return tf.reduce_mean(1.0 - per_row_accuracy, name=p_aba.name[:-2] + '_esterr')
    


import tensorflow as tf
from tensorflow.python.ops import array_ops

def focal_loss (labels, logits, alpha = 0.25, gamma = 2):
    zeros = array_ops.zeros_like (logits, dtype = p.dtype)
    
    # For poitive prediction, only need consider front part loss, back part is 0;
    # target_tensor > zeros <=> z=1, so poitive coefficient = z - p.
    pos_p_sub = array_ops.where (labels > zeros, labels - logits, zeros)
    
    # For negative prediction, only need consider back part loss, front part is 0;
    # target_tensor > zeros <=> z=1, so negative coefficient = 0.
    neg_p_sub = array_ops.where (labels > zeros, zeros, logits)
    per_entry_cross_ent = - alpha * (pos_p_sub ** gamma) * tf.log (tf.clip_by_value (logits, 1e-8, 1.0)) \
                          - (1 - alpha) * (neg_p_sub ** gamma) * tf.log (tf.clip_by_value(1.0 - logits, 1e-8, 1.0))
    return tf.reduce_sum (per_entry_cross_ent, axis = 1)

def sigmoid_focal_loss (labels, logits, alpha = 0.25, gamma = 2):
    r"""Compute focal loss for predictions.
        Multi-labels Focal loss formula:
            FL = -alpha * (z-p)^gamma * log(p) -(1-alpha) * p^gamma * log(1-p)
                 ,which alpha = 0.25, gamma = 2, p = sigmoid(x), z = target_tensor.
    Args:
     logits: A float tensor of shape [batch_size, num_anchors,
        num_classes] representing the predicted logits for each class
     labels: A float tensor of shape [batch_size, num_anchors,
        num_classes] representing one-hot encoded classification targets     
     alpha: A scalar tensor for focal loss alpha hyper-parameter
     gamma: A scalar tensor for focal loss gamma hyper-parameter
    Returns:
        loss: A (scalar) tensor representing the value of the loss function
    """
    return focal_loss (labels, tf.nn.sigmoid (logits), alpha = 0.25, gamma)

def softmax_focal_loss (labels, logits, alpha = 0.25, gamma = 2):
    r"""Compute focal loss for predictions.
        Multi-labels Focal loss formula:
            FL = -alpha * (z-p)^gamma * log(p) -(1-alpha) * p^gamma * log(1-p)
                 ,which alpha = 0.25, gamma = 2, p = sigmoid(x), z = target_tensor.
    Args:
     logits: A float tensor of shape [batch_size, num_anchors,
        num_classes] representing the predicted logits for each class
     labels: A float tensor of shape [batch_size, num_anchors,
        num_classes] representing one-hot encoded classification targets     
     alpha: A scalar tensor for focal loss alpha hyper-parameter
     gamma: A scalar tensor for focal loss gamma hyper-parameter
    Returns:
        loss: A (scalar) tensor representing the value of the loss function
    """
    return focal_loss (labels, tf.nn.softmax (logits), alpha = 0.25, gamma)

'''
def sigmoid_focal_loss (labels, logits, alpha = 0.25, gamma = 2):
    """
    Computer focal loss for binary classification
    Args:
      labels: A int32 tensor of shape [batch_size].
      logits: A float32 tensor of shape [batch_size].
      alpha: A scalar for focal loss alpha hyper-parameter. If positive samples number
      > negtive samples number, alpha < 0.5 and vice versa.
      gamma: A scalar for focal loss gamma hyper-parameter.
    Returns:
      A tensor of the same shape as `lables`
    """
    y_pred = tf.nn.sigmoid (logits)
    labels = tf.to_float (labels)
    L = -labels * (1 - alpha) * ((1 - y_pred) * gamma) * tf.log (y_pred) - \
      (1 - labels) * alpha * (y_pred ** gamma) * tf.log (1 - y_pred)
    return L

def softmax_focal_loss (labels, logits, gamma = 2):
    """
    Computer focal loss for multi classification
    Args:
      labels: A int32 tensor of shape [batch_size].
      logits: A float32 tensor of shape [batch_size,num_classes].
      gamma: A scalar for focal loss gamma hyper-parameter.
    Returns:
      A tensor of the same shape as `lables`
    """
    y_pred = tf.nn.softmax (logits, axis = 1) # [batch_size,num_classes]
    # labels = tf.one_hot (labels, depth = y_pred.shape [1])
    L = -labels * ((1.0 - y_pred) ** gamma) * tf.log (y_pred)
    L = tf.reduce_sum (L, axis = 1)
    return L
'''


import tensorflow as tf


def grad (cost, learning_rate, global_step, **karg):
    return tf.train.GradientDescentOptimizer (learning_rate).minimize (cost, global_step = global_step, **karg)

def momentum (cost, learning_rate, global_step, use_nesterov = False, momentum = 0.9, **karg):    
    return tf.train.MomentumOptimizer (learning_rate, momentum, use_nesterov = use_nesterov).minimize (cost, global_step = global_step, **karg)

def nesterov_momentum (cost, learning_rate, global_step, **karg):
    return momentum (cost, learning_rate, global_step, True, **karg) 
nag = nesterov_momentum

def adagrad (cost, learning_rate, global_step, initial_accumulator_value = 0.1, **karg):
    return tf.train.AdagradOptimizer (learning_rate, initial_accumulator_value).minimize(cost, global_step = global_step, **karg)

def adadelta (cost, learning_rate, global_step, rho = 0.95, **karg):
    return tf.train.AdadeltaOptimizer (learning_rate, rho).minimize(cost, global_step = global_step, **karg)

def rmsprop (cost, learning_rate, global_step, decay = 0.9, momentum = 0.0, **karg):
    return tf.train.RMSPropOptimizer (learning_rate, decay, momentum).minimize(cost, global_step = global_step, **karg)

def adam (cost, learning_rate, global_step, beta1 = 0.9, beta2 = 0.999, **karg):
    return tf.train.AdamOptimizer (learning_rate, beta1, beta2).minimize (cost, global_step = global_step, **karg)

def ftrl (cost, learning_rate, global_step, initial_accumulator_value=0.1, l1_regularization_strength=0.0, l2_regularization_strength=0.0, **karg):
    return tf.train.FtrlOptimizer (learning_rate, global_step, initial_accumulator_value, l1_regularization_strength, l2_regularization_strength).minimize(cost, global_step = global_step, **karg)

def proxadagrad (cost, learning_rate, global_step, initial_accumulator_value=0.1, l1_regularization_strength=0.0, l2_regularization_strength=0.0, **karg):
    return tf.train.ProximalAdagradOptimizer (learning_rate, global_step, initial_accumulator_value, l1_regularization_strength, l2_regularization_strength).minimize(cost, global_step = global_step, **karg)

def proxgrad (cost, learning_rate, global_step, l1_regularization_strength=0.0, l2_regularization_strength=0.0, **karg):
    return tf.train.ProximalAdagradOptimizer (learning_rate, global_step, l1_regularization_strength, l2_regularization_strength).minimize(cost, global_step = global_step, **karg)

def clip (cost, learning_rate, global_step, min_, max_):
    train_op = tf.train.AdamOptimizer (learning_rate = learning_rate)
    gradients = train_op.compute_gradients (cost)
    capped_gradients = [(tf.clip_by_value (grad, min_, max_), var) for grad, var in gradients]
    return train_op.apply_gradients (capped_gradients, global_step = global_step)



#--------------------------------------------------------
# Feature Selectors For Unsupervised Learning
# with TFIDF Vectors
# 2017.11.3 Hans roh
#--------------------------------------------------------

from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.feature_selection import VarianceThreshold, SelectKBest, SelectPercentile, chi2
from sklearn.metrics.pairwise import cosine_similarity
import sklearn.cluster as cl
import numpy as np
import pandas as pd
import pickle
import sys, os
import time
from pprint import pprint as pp

class Variant:		
	def __init__ (self, samples):
		self.samples = samples
		
	def select (self, top_n = 6000, min_tfidf = 0.1, reverse = False):	
		D = self.samples.matrix.toarray ()
		
		scores = np.var (D, axis = 0)
		# shape check
		assert len (scores) == len (D [0])
		
		ids = np.argsort(scores)[::-1]
		if reverse:			
			topn_ids = ids [-top_n:]
		else:
			topn_ids = ids [:top_n]
		
		top_feats = [(self.samples.names [i], i, scores [i]) for i in topn_ids]
		print ("{} features was selected".format (len (top_feats)))
		return top_feats
		

class ClusterMaxPool (Variant):	
	def select (self, top_n = 6000, min_tfidf = 0.1, reverse = False):	
		# https://stats.stackexchange.com/questions/266220/tfidf-for-feature-selection-method-for-unlabeled-text-documents
		D = self.samples.matrix.toarray ()
		covar = np.cov (D, rowvar = False)	
		k_means = cl.KMeans(init='k-means++', n_clusters = top_n, n_init=10)
		k_means.fit (covar)
		
		CS = {}
		for i, c in enumerate (k_means.labels_):
			if c not in CS:
				CS [c] = []			
			CS [c].append (self.samples.names [i])
		pp (CS)





import numpy as np
from skimage import exposure
import cv2

def contrast_stretching (x, r1 = 2, r2 = 98):
    floated = True
    if np.max (x) > 1:
         floated = False
         x = x / 255
    p2, p98 = np.percentile (x, (r1, r2))
    x = exposure.rescale_intensity (x, in_range = (p2, p98))
    if not floated:
         return np.clip ((x * 255).astype ('int64'), 0, 255)
    return x

def adaptive_equalization (x, clip_limit = 0.03):
     return exposure.equalize_adapthist (x, clip_limit = clip_limit)
                
def histogram_equalization (x):
    return exposure.equalize_hist (x)

def rotate (x, angle):
    row, col = x.shape [:2]
    center = tuple (np.array ([row,col]) / 2)
    rot_mat = cv2.getRotationMatrix2D (center, angle, 1.0)
    new_image = cv2.warpAffine(x, rot_mat, (col, row))
    return new_image

def flip (x):
    return cv2.flip(x, 1)
vflip = flip

def hflip (x):
    return cv2.flip(x, 0)    

def whitening (x):
    mean = np.mean (x)    
    std = np.std (x)  
    return (x - mean) / std



#
# Copyright (c) Microsoft. All rights reserved.
# Licensed under the MIT license. See LICENSE.md file in the project root for full license information.
#

import math
  
class Point(object):
    def __init__(self, x=0.0, y=0.0):
        self.x = x
        self.y = y
     
    def __add__(self, p):
        """Point(x1+x2, y1+y2)"""
        return Point(self.x+p.x, self.y+p.y)
     
    def __sub__(self, p):
        """Point(x1-x2, y1-y2)"""
        return Point(self.x-p.x, self.y-p.y)
     
    def __mul__( self, scalar ):
        """Point(x1*x2, y1*y2)"""
        return Point(self.x*scalar, self.y*scalar)
     
    def __div__(self, scalar):
        """Point(x1/x2, y1/y2)"""
        return Point(self.x/scalar, self.y/scalar)
     
    def __str__(self):
        return "(%s, %s)" % (self.x, self.y)
     
    def length(self):
        return math.sqrt(self.x**2 + self.y**2)
     
    def distance_to(self, p):
        """Calculate the distance between two points."""
        return (self - p).length()
     
    def as_tuple(self):
        """(x, y)"""
        return (self.x, self.y)
     
    def clone(self):
        """Return a full copy of this point."""
        return Point(self.x, self.y)
     
    def integerize(self):
        """Convert co-ordinate values to integers."""
        self.x = int(self.x+0.5)
        self.y = int(self.y+0.5)
     
    def floatize(self):
        """Convert co-ordinate values to floats."""
        self.x = float(self.x)
        self.y = float(self.y)
     
    def reset(self, x, y):
        """Reset x & y coordinates."""
        self.x = x
        self.y = y
     
    def shift(self, pt):
        """Move to new (x+pt.x,y+pt.y)."""
        self.x = self.x + pt.x
        self.y = self.y + pt.y
     
    def shift_xy(self, dx, dy):
        """Move to new (x+dx,y+dy)."""
        self.x = self.x + dx
        self.y = self.y + dy
     
    def rotate(self, rad):
        """Rotate counter-clockwise by rad radians.
        Positive y goes *up,* as in traditional mathematics.
        The new position is returned as a new Point.
        """
        s, c = [f(rad) for f in (math.sin, math.cos)]
        x, y = (c*self.x - s*self.y, s*self.x + c*self.y)
        return Point(x,y)
     
    def rotate_about(self, p, theta):
        """Rotate counter-clockwise around a point, by theta degrees.
        Positive y goes *up,* as in traditional mathematics.
        The new position is returned as a new Point.
        """
        result = self.clone()
        result.shift(-p.x, -p.y)
        result.rotate(theta)
        result.shift(p.x, p.y)
        return result
  
class Rect(object):
    """The rectangle stores left, top, right, and bottom values.
    Coordinates are based on screen coordinates.
    origin                            top
    +-----> x increases                |
    |                           left  -+-  right
    v                                  |
    y increases                      bottom
    """

    def __init__(self, box):
        """Initialize a rectangle from two points."""
        self.left = box[0]
        self.top = box[1]
        self.right = box[2]
        self.bottom = box[3]
 
    def as_tuple(self):
        """(left, top, right, bottom)"""
        return (self.left, self.top, self.right, self.bottom)

    def width(self): 
        """Width"""
        return (self.right - self.left) 

    def height(self): 
        """Height"""
        return (self.bottom - self.top) 

    def contains(self, pt):
        """Return true if a point is inside the rectangle."""
        x,y = pt.as_tuple()
        return (self.left <= x <= self.right and
                self.top <= y <= self.bottom)

    def shift(self, pt):
        """Shift by pt.x and pt.y."""
        self.left = self.left + pt.x
        self.right = self.right + pt.x
        self.top = self.top + pt.y
        self.bottom = self.bottom + pt.y

    def shift_xy(self, dx, dy): 
        """Shift by dx and dy."""
        self.left = self.left + dx
        self.right = self.right + dx
        self.top = self.top + dy
        self.bottom = self.bottom + dy
 
    def equal(self, other): 
        """Return true if a rectangle is identical to this rectangle."""
        return (self.right == other.left and self.left == other.right and
                self.top == other.bottom and self.bottom == other.top)
 
    def overlaps(self, other):
        """Return true if a rectangle overlaps this rectangle."""
        return (self.right > other.left and self.left < other.right and
                self.top < other.bottom and self.bottom > other.top)

    def intersect(self, other): 
        """Return the intersect rectangle.
        Note we don't check here whether the intersection is valid
        If needed, call overlaps() first to check 
        """ 
        return Rect((max(self.left, other.left), 
                    max(self.top, other.top), 
                    min(self.right, other.right), 
                    min(self.bottom, other.bottom)))
    
    def clamp(self, xmin, ymin, xmax, ymax): 
        """Return clamped rectangle based on the other rectangle.
        Note we don't check here whether the output is valid
        If needed, call overlaps() first to check 
        """
        self.left = max(self.left, xmin)
        self.right = min(self.right, xmax)
        self.top = max(self.top, ymin)
        self.bottom = min(self.bottom, ymax)         

    def top_left(self):
        """Return the top-left corner as a Point."""
        return Point(self.left, self.top)
     
    def bottom_right(self):
        """Return the bottom-right corner as a Point."""
        return Point(self.right, self.bottom)

    def center(self): 
        """Return the center as a Point.""" 
        return Point((self.left+self.right)/2.0, (self.top+self.bottom)/2.0) 

    def mult(self, xmul, ymul): 
        """Return a rectangle with all coordinates multipled by a number.""" 
        return Rect((self.left*xmul, self.top*ymul, self.right*xmul, self.bottom*ymul))
     
    def scale(self, scale):
        """Return a scaled rectangle with identical center."""
        xctr = (self.left + self.right)/2.0
        yctr = (self.top + self.bottom)/2.0
        width = self.width()*scale
        height = self.height()*scale
        xstart = xctr-width/2.0
        ystart = yctr-height/2.0
        return Rect((xstart, ystart, xstart+width, ystart+height))

    def cocenter(self, new_width, new_height): 
        """Return a new rectangle with identical center."""
        xctr = (self.left + self.right)/2.0
        yctr = (self.top + self.bottom)/2.0
        xstart = xctr - new_width/2.0
        ystart = yctr - new_height/2.0
        return Rect((xstart, ystart, xstart+new_width, ystart+new_height))

    def integerize(self):
        """Convert co-ordinate values to integers."""
        self.left = int(self.left+0.5)
        self.right = int(self.right+0.5)
        self.top = int(self.top+0.5)
        self.bottom = int(self.bottom+0.5)
     
    def floatize(self):
        """Convert co-ordinate values to floats."""
        self.left = float(self.left)
        self.right = float(self.right)
        self.top = float(self.top)
        self.bottom = float(self.bottom)
     
    def __str__( self ):
        return "<Rect (%s,%s)-(%s,%s)>" % (self.left,self.top,
                                        self.right,self.bottom)
     


from pyssml.AmazonSpeech import AmazonSpeech
from pyssml.PySSML import PySSML

class SSML (AmazonSpeech):
    def prosody(self, attributes, word):
        tag_attributes = ''
        if attributes is None:
            raise TypeError('Parameter attributes must not be None')
        if word is None:
            raise TypeError('Parameter word must not be None')
        try:
            for k, v in attributes.items():
                v = v.lower().strip()
                if v in PySSML.PROSODY_ATTRIBUTES[k]:
                    tag_attributes += " %s='%s'" % (k, v)                
                elif k in ('pitch', 'rate'):
                    rate_value = int(''.join([c for c in v if c in '-0123456789']))
                    tag_attributes += " %s='%d%%'" % (k, rate_value)                    
                elif k == 'volume':
                    rate_value = int(''.join([c for c in v if c in '-0123456789']))
                    tag_attributes += " %s='%ddB'" % (k, rate_value)
                else:
                    raise ValueError('Attribute %s value %s is invalid' % (v, k))
            self.ssml_list.append("<prosody%s>%s</prosody>" % (tag_attributes, self._escape(word)))

        except AttributeError:
            raise AttributeError('Parameters must be strings')
        
        except KeyError:
            raise KeyError('Attribute is unknown')
        
        except ValueError:
            raise ValueError('Attribute value is invalid')



import numpy as np
import copy

class Labels:
    def __init__ (self, labels = []):
        self.labels = labels

    def __getattr__ (self, item):        
        if isinstance(item, str) and item[:2] == item[-2:] == '__':
            # skip non-existing dunder method lookups
            raise AttributeError(item)    
        return getattr (self.labels, item)

    def __len__ (self):
        return len (self.labels)

    def __getitem__(self, sliced):
        return self.labels [sliced]

    def __repr__ (self):
        return repr (self.labels)
        
    def add (self, label):
        self.labels.append (label)

    def encode (self, *args):    
        assert len (args) == len (self.labels)
        y = []
        for idx, label in enumerate (self.labels):
            val = args [idx]
            if isinstance (val, dict):
                y = label.setval (val, prefix = y)     
            elif isinstance (val, (list, tuple)):
                y = label.ones (val, prefix = y)                    
            else:
                y = label.onehot (val, prefix = y)    
        return y

    
class Label:
    def __init__ (self, items, name = "label"):
        self._origin = items
        self.name= name
        self._indexes = {}        
        if isinstance (self._origin, (list, tuple)):
            self._set = list (set (self._origin))
            self._set.sort ()                        
        else:
            assert isinstance (items, dict)
            try:
                assert len (self._origin) > 1
                items_ = sorted (self._origin.items (), key = lambda x: x [1])
                pos = [v for k, v in items_]
                self._set = [k for k, v in items_]
                assert len (items_) == len (set (pos))
                assert pos [0] == 0
                assert pos [-1] == len (pos) - 1
            except (TypeError, AssertionError): # val is None
                self._set = sorted (self._origin.keys ())
                
        for idx, item in enumerate (self._set):
            self._indexes [item] = idx
        self._items = dict ([(v, k) for k, v in self._indexes.items ()])        
        
    def __repr__ (self):
        return "<Label ({}): {}>".format (self.name, "[" + ", ".join ([str (each) for each in self._set]) + "]") 
    
    def __getitem__ (self, index):
        return self.item (index) 
        
    def info (self, item):
        return self._origin [item] 
    
    def __len__ (self):
        return len (self._set)
    
    def index (self, item):
        return self._indexes [item]
    
    def item (self, index):
        return self._items [index]
    
    def items (self):
        return self._set
    
    def top_k (self, arr, k = 1):
        items = []         
        for idx in np.argsort (arr)[::-1][:k]:
            items.append (self._items [idx])
        return items
    
    def setval (self, items, type = np.float, prefix = None):
        arr = np.zeros (len (self._set)).astype (type)
        if not isinstance (items, (dict, list, tuple)):
            items = [items]

        for item, value in items.items ():
            tid = self._indexes.get (item, -1)
            if tid == -1:
                raise KeyError ("{} Not Found".format (item))
            arr [self._indexes [item]] = value
            
        if prefix is not None:
            return np.concatenate ([prefix, arr])
        else:
            return arr

    def onehot (self, item, type = np.float, prefix = None):
        return self.setval ({item: 1.0}, type, prefix)
    one = onehot #lower version compatible

    def ones (self, items, type = np.float, prefix = None):
        return self.setval (dict ([(item, 1.0) for item in items]), type, prefix)


def onehots (labels, vals):
    if not isinstance (labels, list):
        labels = [labels]
    if not isinstance (vals, list):
        vals = [vals]
    y = []    
    for idx, label in enumerate (labels):
        y = label.onehot (vals [idx], prefix = y)
    return y

if __name__ == "__main__":    
    v1 = Label (["a", "b", "c", "d", "e"])
    v2 = Label (["a", "b", "c", "d", "e"])
    v3 = Label (["a", "b", "c", "d", "e"])
    f = Labels ([v1, v2, v3])
    print (f.encode ("c", {"a": 0.5}, ['d', 'e']))


    
    
    


import tensorflow as tf
from . import normalizer
import os
import pickle
from tensorflow.python.framework import tensor_util

def check_or_create (sess, graph):
    if sess is None:
        sess = tf.Session ()
    if graph is None:
        graph = sess.graph
    return sess, graph
        
def convert (model_dir, predict_def, inputs, outputs, sess = None, graph = None, train_dir = None):
    sess, graph = check_or_create (sess, graph)
    
    with graph.as_default ():
        builder = tf.saved_model.builder.SavedModelBuilder(model_dir)
        inputs=dict ([(k, tf.saved_model.utils.build_tensor_info (v)) for k, v in inputs.items ()])
        outputs=dict ([(k, tf.saved_model.utils.build_tensor_info (v)) for k,v in outputs.items ()])
        
        prediction_signature = (
          tf.saved_model.signature_def_utils.build_signature_def(
              inputs=inputs,
              outputs=outputs,
              method_name=tf.saved_model.signature_constants.PREDICT_METHOD_NAME)
          )

        builder.add_meta_graph_and_variables(
          sess, 
          [tf.saved_model.tag_constants.SERVING],
          signature_def_map = {
              tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY: prediction_signature,
              predict_def: prediction_signature
          }
        )
        builder.save ()
    return inputs, outputs
 
def load (model_dir, sess = None, graph = None):
    sess, graph = check_or_create (sess, graph)
        
    with graph.as_default ():            
        meta = tf.saved_model.loader.load (
            sess, 
            [tf.saved_model.tag_constants.SERVING], 
            model_dir
        )
    
    input_map = {}
    outputs = {}
    activation = {}
        
    for signature_def_name, signature_def in meta.signature_def.items ():
        input_map [signature_def_name] = {}
        outputs [signature_def_name] = []
        activation [signature_def_name] = []
        
        for k, v in signature_def.inputs.items ():
            input_map [signature_def_name][k] = (v.name, sess.graph.get_tensor_by_name (v.name), v.dtype, [dim.size for dim in v.tensor_shape.dim])
        for k, v in signature_def.outputs.items ():
            outputs[signature_def_name].append ((k, v.name, sess.graph.get_tensor_by_name (v.name), v.dtype, [dim.size for dim in v.tensor_shape.dim]))                
            activation [signature_def_name] .append (v.name)
    
    return Interpreter (model_dir, meta, input_map, outputs, activation, sess, graph)


class Interpreter:
    def __init__ (self, model_dir, meta, input_map, outputs, activation, sess, graph):
        self.model_dir = model_dir
        self.meta = meta
        self.input_map = input_map
        self.outputs = outputs
        self.activation = activation
        self.sess = sess
        self.graph = graph
        self.norm_factor = self.load_norm_factor ()
        
    def load_norm_factor (self):
        norm_file = os.path.join (self.model_dir, "normfactors")
        if os.path.isfile (norm_file):
            with open (norm_file, "rb") as f:
                return pickle.load (f)
    
    def normalize (self, x):
        # SHOLUD BE SYMC with dnn,DNN.normalize ()
        if not self.norm_factor:
            return x
        return normalizer.normalize (x, *self.norm_factor)
    
    def  run (self, x, **kargs):
        feed_dict = {}
        kargs ["x"] = x
        for k, v in kargs.items ():
            tensor_name, tensor, dtype, shape = self.input_map [tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY][k]        
            if k == "x":
                v = self.normalize (v)
            feed_dict [tensor] =  v
        return self._run (feed_dict)
        
    def _run (self, feed_dict, signature_def_name = None):
        if signature_def_name is None:
            signature_def_name = tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY
        return self.sess.run (self.activation [signature_def_name], feed_dict = feed_dict)




# required tensorflow >= 1.9 

from tensorflow.contrib import lite
from tensorflow.python.tools import freeze_graph
import tensorflow as tf
import numpy as np
import threading
from tensorflow.contrib.lite.python.convert import tensor_name
from . import saved_model
import os

"""
freeze_graph --input_checkpoint=resources/chkpoint/0025-acc-87.267+cost-2.988-43750 \
    --input_graph=resources/exported/toco/graph-def.pb \
    --input_binary=true \
    --output_node_names=concat_1:0 \
    --output_graph=resources/exported/toco/frozen-graph.pb

toco --graph_def_file=resources/exported/lge7/70/frozen-graph.pb \
  --output_format=TFLITE \
  --output_file=resources/exported/lge7/70/model.tflite \
  --input_arrays=Placeholder \
  --output_arrays=concat_1 \
  --inference_type=QUANTIZED_UINT8 \
  --inference_input_type=QUANTIZED_UINT8 \
  --std_dev_values=128 \
  --mean_values=128 \
  --default_ranges_min=-1 \
  --default_ranges_max=6 \
  --allow_nudging_weights_to_use_fast_gemm_kernel

toco --graph_def_file=resources/exported/lge7/70/frozen-graph.pb \
  --output_format=TFLITE \
  --output_file=resources/exported/lge7/70/model.tflite \
  --input_arrays=Placeholder \
  --output_arrays=concat_1
"""

def convert (path, saved_model_dir, quantize = False, quantized_input_stats = (128, 128), default_ranges_stats = (0, 1)):
    try:
        converter = tf.contrib.lite.TFLiteConverter.from_saved_model (saved_model_dir)
    except AttributeError:
        converter = tf.contrib.lite.TocoConverter.from_saved_model (saved_model_dir)
    if quantize:
        converter.inference_type = tf.contrib.lite.constants.QUANTIZED_UINT8
        converter.quantized_input_stats = {tensor_name (converter._input_tensors [0]): quantized_input_stats}
        converter.default_ranges_stats =  default_ranges_stats
        converter.allow_nudging_weights_to_use_fast_gemm_kernel = True
                    
    tflite_model = converter.convert()
    open(path, "wb").write(tflite_model)


def convert_from_graph_def (train_dir, graph_def_file, input_arrays, output_arrays, quantize = False, quantized_input_stats = (128, 128), default_ranges_stats = (0, 1)):
    checkpoints = [] 
    for f in os.listdir (train_dir):
        s = f.find (".data-")
        if s == -1:
            continue
        checkpoints.append (f [:s])
    
    checkpoints.sort (key = lambda x: int (x.split ("-")[-1]))
    output_graph = os.path.join (os.path.dirname (graph_def_file), "frozen-graph-def.pb")
    output_model = os.path.join (os.path.dirname (graph_def_file), "model.tflite")
    
    cmd = [
        "freeze_graph --input_checkpoint={}".format (os.path.join (train_dir, checkpoints [-1])),
        "--input_graph={} --input_binary=true".format (graph_def_file),
        "--output_node_names={}".format (",".join (output_arrays)),
        "--output_graph={}".format (output_graph)        
    ]
    os.system (" ".join (cmd))
    
    cmd = [    
      "tflite_convert --graph_def_file={}".format (output_graph),
      "--output_format=TFLITE",
      "--output_file={}".format (output_model),      
      "--input_arrays={}".format (".".join (input_arrays)),
      "--output_arrays={}".format (".".join (output_arrays))
    ]
    if quantize:
        cmd.append ("--inference_type=QUANTIZED_UINT8")
        cmd.append ("--inference_input_type=QUANTIZED_UINT8")
        cmd.append ("--mean_values={}".format (quantized_input_stats [0]))
        cmd.append ("--std_dev_values={}".format (quantized_input_stats [1]))
        cmd.append ("--default_ranges_min={}".format (default_ranges_stats [0]))
        cmd.append ("--default_ranges_max={}".format (default_ranges_stats [1]))
        cmd.append ("--allow_nudging_weights_to_use_fast_gemm_kernel=true")   

    os.system (" ".join (cmd))
    
def load (path, quantized = None):
    return Interpreter (path, quantized)


class Interpreter (saved_model.Interpreter):    
    def __init__ (self, path, quantized = None, debug = False):        
        self.path = path
        self._quantized = quantized
        self.debug = debug
        self.interp = lite.Interpreter (path)
        self.model_dir = os.path.dirname (path)
        self.norm_factor = self.load_norm_factor ()
        self.input = self.interp.get_input_details()[0]["index"]
        self.output = self.interp.get_output_details()[0]["index"]
        self.lock = threading.RLock ()
        if self.is_quantized ():
            assert self._quantized and len (self._quantized) == 2, "quantized should be 2 integer tuple like (128, 128)"
        else:
            self._quantized = None
        self._allocated = False
        
    def is_quantized (self):
        return self.interp.get_input_details()[0]["dtype"] is np.uint8
        
    def get_info (self): 
        inputs = {}
        outputs = {}
        for each in self.interp.get_input_details ():
            inputs [each ["name"]] = (each ["index"], each ["shape"], each ["dtype"])
        for each in self.interp.get_output_details ():
            outputs [each ["name"]] = (each ["index"], each ["shape"], each ["dtype"])     
        return inputs, outputs      
        
    def run (self, x, **kargs):
        if not self._allocated:
            if self._quantized:
                self.interp.resize_tensor_input (self.input, np.array(x.shape, dtype=np.int32))
            self.interp.allocate_tensors ()
            self._allocated = True
        
        if self.debug:
            print ("unorm", x [0][0])
        x =  self.normalize (x)
        if self.debug:
            print ("normed", x [0][0])
        if self._quantized:
            x = x * self._quantized [0] + self._quantized [1]
            x = np.clip (x, 0, 255).astype ("uint8")            
            if self.debug:
                print ("qunatized", x [0][0])
        else:
            x = x.astype ("float32")
        
        with self.lock:
            self.interp.set_tensor (self.input, x)
            self.interp.invoke ()
            y = self.interp.get_tensor (self.output)
        return y



from .dnn import DNN
from .multidnn import MultiDNN  

__version__ = "0.3.4.27"

def subprocess (train_func):
    from multiprocessing import Pool    
    with Pool(1) as p:
        return p.apply (train_func, ())
    



import librosa
import soundfile
import os
from rs4 import pathtool
from ..video.ffmpeg import run_process 

def resampling (audioFilePath, targetFilePath, sampling_rate = 16000):
    y, sr = librosa.load (audioFilePath, sr = 22050, mono = True)
    with open (targetFilePath, "wb") as f:
       soundfile.write (f, y, sampling_rate)
       
def resample_dir (directory, target = None, recusive = False):
    if taget is None:
        target  = os.path.join (directory, "AIMDV")
    pathtool.mkdir (target)
        
    for each in os.listdir (directory):
        if each == "AIMDV":
            continue
        
        path = os.path.join (directory, each)
        if os.path.isdir (path):
            if recusive:
                resample_dir (path, target, True)
            else:
                continue    
        
        try:
            resampling (path, target)
        except:
            raise

def convert2mp3 (input, output, quality = 7):
    try:
        int (quality)
    except ValueError:
        assert quality.endswith ("k")
        q = "-b:a {}".format (quality)
    else:
        q = "-q:a {}".format (quality)            
    os.path.isfile (output) and os.remove (output)
    cmd = 'ffmpeg -i "{}" -codec:a libmp3lame {} "{}"'.format (input, q, output)
    r = run_process (cmd)
    assert os.path.isfile (output)
    return output
    
def convert2wav (input, output, sample_rate = 44050):
    os.path.isfile (output) and os.remove (output)        
    cmd = 'ffmpeg -i "{}" -acodec pcm_s16le -ar {} "{}"'.format (input, sample_rate, output)
    r = run_process (cmd)
    #r = run_process ('ffmpeg -i "{}" "{}"'.format (input, output))
    assert os.path.isfile (output)
    return output
    
    


import numpy as np
from rs4 import pathtool
from hashlib import md5
import os
import math
import rs4
import random

class DataSet:
    def __init__ (self, augment = 0, shuffle = False):
        self.data = []        
        self.augment = augment
        self.shuffle = shuffle
        self.cache = {}     
        self.xs, self.ys = None, None

    def __len__ (self):
        return len (self.data)
        
    def add (self, row):
        self.data.append (row)

    def get_cache_name (self, path):
        return md5 (path.encode ()).hexdigest () + '.{}.npy'.format (self.augment and 'b' or 'a')

    def get_abs_cache_name (self, path):
        return md5 (path.encode ()).hexdigest () + '.npy'
    
    def next_indexes (self, size):
        if self.ys is None:
            self.xs, self.ys = self.initiate_data ()
            self.indexes = np.random.permutation (len (self.data))
            self.current_index = 0

        if not size:
            batch_indexes = self.indexes            
        else:
            size = min (size, len (self.data))
            batch_indexes = list (self.indexes [self.current_index:self.current_index + size])
            if len (batch_indexes) < size:
                self.indexes = np.random.permutation (len (self.data))
                self.current_index = size - len (batch_indexes)
                batch_indexes += list (self.indexes [:self.current_index])
            else:
                self.current_index = self.current_index + size
        return batch_indexes

    # override --------------------------------------------
    def initiate_data (self):
        # reconstruct self.data and return xs, ys for iteration
        raise NotImplementedError

    def next_minibatch (self, size = 0):
        batch_indexes = self.next_indexes (size)
        





import numpy as np
import sys
import os
import glob
import wave
import multiprocessing
from functools import partial
from scipy.fftpack import fft
from scipy.io import wavfile
from scipy import signal
from tqdm import tqdm
import librosa
import librosa.util
import random 
from scipy.stats import skew, kurtosis
import warnings
import resampy

SAMPLE_RATE = 16000
TIME_LAP = 0.2

def _save (data, wavfile, target_path):
    savedir = os.path.join (target_path, os.path.basename (os.path.dirname (os.path.dirname (wavfile))))
    if not os.path.isdir (savedir):
        os.mkdir (savedir)
    target_file = os.path.join(savedir, "%s.npy" % os.path.basename (wavfile))
    np.save(target_file, data)

def norm_volmue (y):
    return y / np.max (np.abs (y))
_normalize = norm_volmue

def whitening (y):
    mean = np.mean (y)    
    std = np.std (y)  
    return (y - mean) / std
    
def log (x, name = ""):
    return np.log (np.abs (x) + 1e-3) / 6.7
  
def add_padding (x, padding):
    if not padding:
        return x
    return np.concatenate ([x, np.zeros (padding)])

def to_db (S):
  S = np.abs (S) + 1e-10 
  return 10 * np.log10(S) - 10 * np.log10(np.max (S))

def gain (input, output, target_dBFS = -20.0):
    from pydub import AudioSegment
    
    sound = AudioSegment.from_file (input)
    change_in_dBFS = target_dBFS - sound.dBFS
    normalized_sound = sound.apply_gain (change_in_dBFS)
    normalized_sound.export(output, output.split (".")[-1].lower ())
    return normalized_sound.dBFS
        
def _compress (data, name, padding = 0, valid_max = 2.0, valid_min = -2.0):
    # padding 3 means 2 convolution, 1 polling    
    data = data.real
    if name == "poly_features":
        data = data / 10.
    elif name == "rmse":
        data = data / 4.
    elif name == "spectral_contrast":
        data = data / 50.
    elif name == "mel":
        data = data / -80.
    elif name in ("stft", "spectral_centroid", "spectral_bandwidth", "spectral_rolloff"):        
        data = log (data, name)
                          
    features = np.array ([
        add_padding (np.mean (data, axis = 1), padding),
        add_padding (np.max (data, axis = 1), padding),
        add_padding (np.min (data, axis = 1), padding),
        add_padding (np.median (data, axis = 1), padding),
        add_padding (np.var (data, axis = 1), padding),
        add_padding (skew (data, axis = 1) / 15., padding),
        add_padding (np.log (abs (kurtosis (data, axis = 1)) + 1e-3) / 6.7, padding)
    ])
    if np.min (features) < valid_min or np.max (features) > valid_max:
        #print (name, int (np.max (features)), int (np.max (features)))
        raise AssertionError
    return features

def _gen_colored_noise (spectral_shape, samples_to_generate):
    # https://www.kaggle.com/holzner/generating-different-colors-of-noise
    def normalize(samples):
        std = samples.std()
        if std > 0:
            return samples / std
        else:
            return samples
    
    flat_spectrum = np.random.normal(size = samples_to_generate // 2 + 1) + \
            1j * np.random.normal(size = samples_to_generate // 2 + 1)
    return normalize(np.fft.irfft (flat_spectrum * spectral_shape).real)

COLORS = [
    'white',   # flat spectrum
    'pink',    # -3dB (factor 0.5)  per octave / -10 dB (factor 0.1) per decade
    'blue',    # +3dB (factor 2)    per octave / +10 dB (factor 10) per decade
    'brown',   # -6dB (factor 0.25) per octave / -20 dB (factor 0.01) per decade
    'violet'
]
NOISE_CACHE = {}
def _gen_noise (color, samples_to_generate):
    if color == 'white':
        return np.random.normal(size = samples_to_generate)    
    spectrum_len = samples_to_generate // 2 + 1
    if color == 'pink':
        return _gen_colored_noise(1. / (np.sqrt(np.arange(spectrum_len) + 1.)), samples_to_generate)        
    elif color == 'blue':
        return _gen_colored_noise(np.sqrt(np.arange(spectrum_len)), samples_to_generate)    
    elif color == 'brown' or color == 'red':
        return _gen_colored_noise(1. / (np.arange(spectrum_len) + 1), samples_to_generate)
    elif color == 'violet' or color == 'purple':
        return _gen_colored_noise(np.arange(spectrum_len), samples_to_generate)    

for color in COLORS:    
    NOISE_CACHE [color] = _gen_noise (color, 16000 * 180)

def gen_noise (color, samples_to_generate):
    global NOISE_CACHE

    assert samples_to_generate % 2 == 0
    if samples_to_generate == len (NOISE_CACHE [color]):
        return NOISE_CACHE [color]
    if samples_to_generate <= len (NOISE_CACHE [color]):
        start = random.randrange (0, len (NOISE_CACHE [color]) - samples_to_generate)
        return NOISE_CACHE [color][start:start + samples_to_generate]
    return  _gen_noise (color, samples_to_generate)

def augment (y, sr, time_stretch = False, pitch_shift = False, random_laps = False, add_noise = False):    
    if add_noise:
        if y.shape [0] % 2 == 1:
            y = y [:-1] # make even
    
        noise = gen_noise (random.choice (COLORS), y.shape [0]) / random.randrange (10, 50)
        if random.randrange (3) == 0:
            noise += gen_noise (random.choice (COLORS), y.shape [0]) / random.randrange (10, 50)
            noise /= 2

        gain = np.random.uniform(0.01, 0.12)
        #print (color, gain)
        y = y + ((gain * noise).astype ("float32"))
    
    if time_stretch:
        rate = 1.0
        if random.randrange (2):
            rate += max (0.02, random.random () / 10.)
        else:
            rate -= max (0.02, random.random () / 10.)     
        y = librosa.effects.time_stretch (y, rate)
        #print ("time_stretch", rate)
    
    if pitch_shift:
        y = librosa.effects.pitch_shift (y, sr, random.random () / 2.0 * random.choice ([1, -1]), bins_per_octave = 12)        
    
    if random_laps and librosa.get_duration(y, sr = sr) > 1.6:
       time_lap = random.random () / 3
       removable = int (sr * time_lap)
       y = y [removable:len (y) - removable]
       y = np.roll (y, random.randrange (len (y)))
    
    return y   

def augment_with_norm (y, sr = SAMPLE_RATE, *args, **karg):
    y = augment (y, sr, *args, **karg)
    y = norm_volmue (y)
    return whitening (y)

def loadraw (wavfile, sample_rate = SAMPLE_RATE, time_lap = 0.2, due_limit = (0.5, 10), res_type = "kaiser_best"):
    if res_type == "hann_window":
        y, sr_orig = librosa.load (wavfile, None, True)
        y = resampy.resample(y, sr_orig, sample_rate, filter = 'sinc_window', window = signal.hann)
        sr = sample_rate
    else:
        y, sr = librosa.load (wavfile, sample_rate, True, res_type = res_type)
    
    a_duration = librosa.get_duration(y, sr = sr)
    removable = int (sample_rate * time_lap)
    y = y [removable:len (y) - removable]
    # Trim the beginning and ending silence
    try:
        y, index = librosa.effects.trim (y)
    except ValueError:        
        return None, sr   
    duration = librosa.get_duration(y, sr = sr)
    #print ("    + duration: {:2.1f} sec => {:2.1f} sec".format (a_duration, duration))
    if duration < due_limit [0] or duration > due_limit [1]:
        warnings.warn ("audio file length is invalid, ignored")
        return None, sr
    if len (y) % 2 == 1:
        y = y [:-1]
    return y, sr
    
def load (wavfile, sample_rate = SAMPLE_RATE, time_lap = 0.2, due_limit = (1, 5), res_type = "kaiser_best", *args, **karg):
    y, sr = loadraw (wavfile, sample_rate, time_lap, due_limit, res_type)
    if y is None:
        return y, sr    
    return featurize (y, sr)    
 
DEFAULT_RECIPE = [
    "mel", "mfcc", "chroma_cqt", "chroma_cens", "zero_crossing_rate", "tonnetz", "rmse", 
    "spectral_contrast", "poly_features", "chroma_stft", 
    "spectral_centroid", "spectral_bandwidth", "spectral_rolloff"
]
   
def generate (y, sample_rate = SAMPLE_RATE, recipe = DEFAULT_RECIPE, n_mfcc = 40, n_mel = 40, padding = 0):
    feature_stack = []
    numfeat = []
    
    # stft ----------------------------------------------------
    stft = librosa.stft (y, n_fft=1024, win_length=512, hop_length=256)
    if "mel" in recipe:
        # mel specrogram ------------------------------------------
        D = np.abs (stft) ** 2
        S = librosa.feature.melspectrogram(S = D, n_mels = n_mel)
        vec = librosa.power_to_db (S, ref = np.max)
        feature_stack.extend (_compress (vec, "mel", padding))
        numfeat.append (vec.shape [0])        
        
    if "stft" in recipe:
        feature_stack.extend (_compress (vec, "stft", padding))
        numfeat.append (vec.shape [0])
        
    if "mfcc" in recipe:
        vec = librosa.feature.mfcc(S = stft, sr = sample_rate, n_mfcc = n_mfcc, n_fft=512, hop_length = 256)
        feature_stack.extend (_compress (vec, "mfcc", padding))
        numfeat.append (vec.shape [0])
        
    if "chroma_cqt" in recipe:        
        vec = librosa.feature.chroma_cqt (y=y, sr = sample_rate, n_chroma = 12, hop_length = 256)
        feature_stack.extend (_compress (vec, "chroma_cqt", padding))
        numfeat.append (vec.shape [0])
        
    if "chroma_cens" in recipe:
        vec = librosa.feature.chroma_cens (y=y, sr = sample_rate, n_chroma = 12, hop_length = 256)
        feature_stack.extend (_compress (vec, "chroma_cens", padding))    
        numfeat.append (vec.shape [0])
        
    if "zero_crossing_rate" in recipe:
        vec = librosa.feature.zero_crossing_rate (y = y)
        feature_stack.extend (_compress (vec, "zero_crossing_rate", padding))
        numfeat.append (vec.shape [0])
        
    if "tonnetz" in recipe:
        vec = librosa.feature.tonnetz (y=y, sr = sample_rate)
        feature_stack.extend (_compress (vec, "tonnetz", padding))
        numfeat.append (vec.shape [0])
        
    if "rmse" in recipe:
        vec = librosa.feature.rmse (S = stft, hop_length = 256)
        feature_stack.extend (_compress (vec, "rmse", padding))
        numfeat.append (vec.shape [0])
        
    for ff in [each for each in recipe if each in ("spectral_contrast", "poly_features", "chroma_stft", "spectral_centroid", "spectral_bandwidth", "spectral_rolloff")]:
        vec = getattr (librosa.feature, ff) (y = y, sr = sample_rate, hop_length = 256)
        feature_stack.extend (_compress (vec, ff, padding))        
        numfeat.append (vec.shape [0])    
    features = np.hstack (tuple (feature_stack))
    assert features.shape == ((sum (numfeat) + len (numfeat) * padding) * 7,)
    return features

AUDIO_ERRS = 0
def make (wavfile, sample_rate = SAMPLE_RATE, time_lap = 0.2, due_limit = (1, 5), recipe = DEFAULT_RECIPE, padding = 0, time_stretch = False, pitch_shift = False, random_laps = False, add_noise = False):
    global AUDIO_ERRS
    
    y, sr = load (wavfile, sample_rate, time_lap, due_limit, time_stretch, pitch_shift, random_laps, add_noise)
    if y is None:
        return
    try:
        return generate (y, sr, recipe, padding = padding)
    except AssertionError:
        AUDIO_ERRS += 1
        print ("audio file may be irregular, ignored ({})".format (AUDIO_ERRS))
        return None    

def save (wavfile, target_path, sample_rate = SAMPLE_RATE, time_lap = 0.2, due_limit = (1, 5), recipe = DEFAULT_RECIPE, padding = 0, **karg):
    features = make (wavfile, sample_rate, time_lap, due_limit, recipe, padding)
    if features is None:
        return False
    _save (features, wavfile, target_path)
    return True

def puff (wavfile, target_path, sample_rate = SAMPLE_RATE, time_lap = 0.2, due_limit = (1, 5), recipe = DEFAULT_RECIPE, padding = 0, n_gen = 3):
    if not os.path.isfile (os.path.join (target_path, wavfile + ".npy")):    
        if not save (wavfile, target_path, sample_rate, time_lap, due_limit, recipe, padding):
            return 0
         
    n = 1
    for i in range (n_gen * 2):
        fn = "{}.{}".format (wavfile, n)
        if os.path.isfile (os.path.join (target_path, fn + ".npy")):
            n += 1
            continue
        params = (random.randrange (2), random.randrange (2), random.randrange (2), random.randrange (2))
        if sum (params) == 0:
            continue
        features = make (wavfile, sample_rate, time_lap, due_limit, recipe, padding, time_stretch = params [0], pitch_shift = params [1], random_laps = params [2], add_noise = params [3], padding = padding)
        if features is None:
            continue        
        _save (features, fn, target_path)
        n += 1
        if n == n_gen:
            break
    
    return n    


if __name__ == '__main__':    
    y, sr = load  (os.path.join (os.path.dirname (__file__), "test.wav"))
    print (y.shape)   
    print (librosa.get_duration(y)) 
    stft = librosa.stft (y [:sr], n_fft=2048, win_length=1200, hop_length=256)
    print (stft.shape)
    ft = generate (y, sr)
    print (ft.shape)
    
    



# https://github.com/hwalsuklee/tensorflow-mnist-VAE

import matplotlib
matplotlib.use('Agg')
import numpy as np
import matplotlib.pyplot as plt
from imageio import imwrite
from skimage.transform import resize as imresize

from rs4 import pathtool
import shutil, os

class Plot_Reproduce_Performance():
    def __init__(self, DIR, n_img_x=8, n_img_y=8, img_w=28, img_h=28, resize_factor=1.0):
        self.DIR = DIR
        pathtool.mkdir (DIR)
        assert n_img_x > 0 and n_img_y > 0

        self.n_img_x = n_img_x
        self.n_img_y = n_img_y
        self.n_tot_imgs = n_img_x * n_img_y

        assert img_w > 0 and img_h > 0

        self.img_w = img_w
        self.img_h = img_h

        assert resize_factor > 0

        self.resize_factor = resize_factor
    
    def reset (self):
        if os.path.isdir (self.DIR):
            shutil.rmtree (self.DIR)
        pathtool.mkdir (self.DIR)
        
    def add_noise (self, images):        
        images = images * np.random.randint (2, size = images.shape)
        images += np.random.randint(2, size = images.shape)
        return images
        
    def save_images(self, images, name='result.jpg'):
        images = images.reshape(self.n_img_x*self.n_img_y, self.img_h, self.img_w)
        imwrite(self.DIR + "/"+name, self._merge(images, [self.n_img_y, self.n_img_x]))

    def _merge(self, images, size):
        h, w = images.shape[1], images.shape[2]

        h_ = int(h * self.resize_factor)
        w_ = int(w * self.resize_factor)

        img = np.zeros((h_ * size[0], w_ * size[1]))

        for idx, image in enumerate(images):
            i = int(idx % size[1])
            j = int(idx / size[1])
            image_ = imresize(image, (w_,h_),  mode = 'reflect')
            img[j*h_:j*h_+h_, i*w_:i*w_+w_] = image_

        return img
    
    
class Plot_Manifold_Learning_Result(Plot_Reproduce_Performance):
    def __init__(self, DIR, n_img_x=20, n_img_y=20, img_w=28, img_h=28, resize_factor=1.0, z_range=2.0):
        self.DIR = DIR

        assert n_img_x > 0 and n_img_y > 0

        self.n_img_x = n_img_x
        self.n_img_y = n_img_y
        self.n_tot_imgs = n_img_x * n_img_y

        assert img_w > 0 and img_h > 0

        self.img_w = img_w
        self.img_h = img_h

        assert resize_factor > 0

        self.resize_factor = resize_factor

        assert z_range > 0
        self.z_range = z_range

        self._set_latent_vectors()

    def _set_latent_vectors(self):

        # z1 = np.linspace(-self.z_range, self.z_range, self.n_img_y)
        # z2 = np.linspace(-self.z_range, self.z_range, self.n_img_x)
        #
        # z = np.array(np.meshgrid(z1, z2))
        # z = z.reshape([-1, 2])

        # borrowed from https://github.com/fastforwardlabs/vae-tf/blob/master/plot.py
        z = np.rollaxis(np.mgrid[self.z_range:-self.z_range:self.n_img_y * 1j, self.z_range:-self.z_range:self.n_img_x * 1j], 0, 3)
        # z1 = np.rollaxis(np.mgrid[1:-1:self.n_img_y * 1j, 1:-1:self.n_img_x * 1j], 0, 3)
        # z = z1**2
        # z[z1<0] *= -1
        #
        # z = z*self.z_range

        self.z = z.reshape([-1, 2])

    # borrowed from https://github.com/ykwon0407/variational_autoencoder/blob/master/variational_bayes.ipynb
    def save_scattered_image(self, z, id, name='scattered_image.jpg'):
        N = 10
        plt.figure(figsize=(8, 6))
        plt.scatter(z[:, 0], z[:, 1], c=np.argmax(id, 1), marker='o', edgecolor='none', cmap=discrete_cmap(N, 'jet'))
        plt.colorbar(ticks=range(N))
        axes = plt.gca()
        axes.set_xlim([-self.z_range-2, self.z_range+2])
        axes.set_ylim([-self.z_range-2, self.z_range+2])
        plt.grid(True)
        plt.savefig(self.DIR + "/" + name)


# borrowed from https://gist.github.com/jakevdp/91077b0cae40f8f8244a
def discrete_cmap(N, base_cmap=None):
    """Create an N-bin discrete colormap from the specified input map"""

    # Note that if base_cmap is a string or None, you can simply do
    #    return plt.cm.get_cmap(base_cmap, N)
    # The following works for string, None, or a colormap instance:

    base = plt.cm.get_cmap(base_cmap)
    color_list = base(np.linspace(0, 1, N))
    cmap_name = base.name + str(N)
    return base.from_list(cmap_name, color_list, N)


import cv2
from rs4 import attrdict

def capture (path, interval = 10):
    count = 0
    vidcap = cv2.VideoCapture (path)
    latest = 0
    frames = []
    while(vidcap.isOpened ()):
        ret, image = vidcap.read ()
        frame_no =  int (vidcap.get(1))
        if latest == frame_no:
            break
        latest = frame_no         
        if frame_no % interval == 0:
            #print('Saved frame number : ' + str(int(vidcap.get(1))))
            frames.append (image)
            #cv2.imwrite("./images/frame%d.jpg" % count, image)
            #print('Saved frame%d.jpg' % count)
            count += 1
    vidcap.release ()
    return frames

def get_info (path):
    vidcap = cv2.VideoCapture (path)
    d = attrdict.AttrDict ()
    d.frame_rate = vidcap.get (cv2.CAP_PROP_FPS)
    d.frame_width = vidcap.get (cv2.CAP_PROP_FRAME_WIDTH)
    d.frame_height = vidcap.get (cv2.CAP_PROP_FRAME_HEIGHT)
    d.video_length = vidcap.get (cv2.CAP_PROP_FRAME_COUNT)  / d.frame_rate    
    d.format = vidcap.get (cv2.CAP_PROP_FORMAT)    
    d.codec = vidcap.get (cv2.CAP_PROP_FOURCC )
    return d



import numpy as np
import math

def mfcc (melspec, n_mfccs):
  # mfcc for android  
  nmel, frames = melspec.shape        
  
  m = math.sqrt(2.0 / nmel)
  DCTcoeffs = np.zeros((nmel, n_mfccs))
  
  a = np.arange (0.5, nmel, 1.0).reshape ([-1, 1])
  b = np.arange (1, n_mfccs + 1, 1.0).reshape ([1, -1])
  DCTcoeffs = a.dot (b)
  DCTcoeffs = m * np.cos (DCTcoeffs * math.pi / nmel)
    
  mfccs = np.zeros ((frames, n_mfccs))
  DCTcoeffs_ = DCTcoeffs.swapaxes (1, 0)
  
  # convert from dB to plain old log magnitude
  melspec_ = melspec / 10.
  for frm in range (frames):
      mfccs [frm] +=  DCTcoeffs_.dot (melspec_ [:,frm]) / frames
  return mfccs.swapaxes(1, 0);



import re
import string

class NGram:
    def __init__ (self, t):
        self._alnums, self.t = self.filter (t)        
        
    rx_splt = re.compile ('[\s%s]' % re.escape('!"#$%&\'()*+,:;<=>?@[\\]^_`{|}~'))        
    rx_force_space = re.compile('(?P<alnum>[0-9a-z]+)', re.I)
    
    def filter (self, t):
        t = self.rx_force_space.sub (" \g<alnum> ", t)
        alnums = []
        for each in self.rx_splt.split (t):
            if self.rx_alnum.search (each):
                alnums.append (each)
        replaced = re.sub ("(%s)" % "|".join (alnums), '', t)
        return [each for each in alnums if not each.isdigit ()], replaced
    
    def alnums (self):
        return self._alnums
    
    def ngram (self, n = 2):        
        sentance = []
        for phrase in self.make_phrases (n):                
            for each in zip (*[phrase[i:] for i in range (n)]):
                each = list (each)
                if n == 2 and (each [0] == " " or each [-1] == " "):
                    continue                
                if each [0] == " ":
                    each [0] = "<"
                elif each [-1] == " ":
                    each [-1] = ">"    
                sentance.append ("".join (each))
        if n == 2:
            return sentance [:-1]
        return sentance    
    
    rx_punct = re.compile('[%s]' % re.escape(string.punctuation))
    rx_space = re.compile('\s+')
    rx_alnum = re.compile('[0-9a-z]{2,}', re.I)
        
    def preprocess (self, t):
        return [" " + self.rx_space.sub (" ", each).strip () + " " for each in self.rx_punct.split (t)]                
    
    def make_phrases (self, n):
        return self.preprocess (self.t)
    
    def analyze (self):
        return self.alnums () + self.ngram (2) + self.ngram (3)


class NGramV2 (NGram):
    # V2: if n is 2, ignoring all spaces    
    def preprocess_nospace (self, t):
        return [self.rx_space.sub ("", each).strip () for each in self.rx_punct.split (t)]

    def make_phrases (self, n = 2):        
        if n == 2:
            return self.preprocess_nospace (self.t)
        return self.preprocess (self.t)
        


#-------------------------------------------------------------------------------
# Original source from: 
#   https://github.com/tensorlayer/tensorlayer/blob/master/tensorlayer/cost.py
#-------------------------------------------------------------------------------


import logging
import tensorflow as tf

def cross_entropy(output, target, name=None):    
    """Softmax cross-entropy operation, returns the TensorFlow expression of cross-entropy for two distributions, it implements
    softmax internally. See ``tf.nn.sparse_softmax_cross_entropy_with_logits``.

    Parameters
    ----------
    output : Tensor
        A batch of distribution with shape: [batch_size, num of classes].
    target : Tensor
        A batch of index with shape: [batch_size, ].
    name : string
        Name of this loss.

    Examples
    --------
    >>> ce = tl.cost.cross_entropy(y_logits, y_target_logits, 'my_loss')

    References
    -----------
    - About cross-entropy: `<https://en.wikipedia.org/wiki/Cross_entropy>`__.
    - The code is borrowed from: `<https://en.wikipedia.org/wiki/Cross_entropy>`__.

    """
    # try: # old
    #     return tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(logits=output, targets=target))
    # except: # TF 1.0
    return tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=target, logits=output, name=name))

_FLOATX = tf.float32
_EPSILON = 1e-10
def categorical_crossentropy (logits, labels):
    """ Categorical Crossentropy.
    Computes cross entropy between y_pred (logits) and y_true (labels).
    Measures the probability error in discrete classification tasks in which
    the classes are mutually exclusive (each entry is in exactly one class).
    For example, each CIFAR-10 image is labeled with one and only one label:
    an image can be a dog or a truck, but not both.
    `y_pred` and `y_true` must have the same shape `[batch_size, num_classes]`
    and the same dtype (either `float32` or `float64`). It is also required
    that `y_true` (labels) are binary arrays (For example, class 2 out of a
    total of 5 different classes, will be define as [0., 1., 0., 0., 0.])
    Arguments:
        y_pred: `Tensor`. Predicted values.
        y_true: `Tensor` . Targets (labels), a probability distribution.
    """
    logits /= tf.reduce_sum(logits,
                            reduction_indices = len (logits.get_shape()) - 1,
                            keepdims=True)
    # manual computation of crossentropy
    logits = tf.clip_by_value(logits, tf.cast(_EPSILON, dtype=_FLOATX),
                              tf.cast(1.-_EPSILON, dtype=_FLOATX))
    cross_entropy = - tf.reduce_sum(labels * tf.log(logits),
                           reduction_indices=len(logits.get_shape())-1)
    return tf.reduce_mean(cross_entropy)


def sigmoid_cross_entropy(output, target, name=None):
    """Sigmoid cross-entropy operation, see ``tf.nn.sigmoid_cross_entropy_with_logits``.

    Parameters
    ----------
    output : Tensor
        A batch of distribution with shape: [batch_size, num of classes].
    target : Tensor
        A batch of index with shape: [batch_size, ].
    name : string
        Name of this loss.

    """
    # try: # TF 1.0
    return tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(labels=target, logits=output, name=name))
    # except:
    #     return tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits=output, targets=target))


def binary_cross_entropy(output, target, epsilon=1e-8, name='bce_loss'):
    """Binary cross entropy operation.

    Parameters
    ----------
    output : Tensor
        Tensor with type of `float32` or `float64`.
    target : Tensor
        The target distribution, format the same with `output`.
    epsilon : float
        A small value to avoid output to be zero.
    name : str
        An optional name to attach to this function.

    References
    -----------
    - `ericjang-DRAW <https://github.com/ericjang/draw/blob/master/draw.py#L73>`__

    """
    #     from tensorflow.python.framework import ops
    #     with ops.op_scope([output, target], name, "bce_loss") as name:
    #         output = ops.convert_to_tensor(output, name="preds")
    #         target = ops.convert_to_tensor(targets, name="target")
    with tf.name_scope(name):
        return tf.reduce_mean(tf.reduce_sum(-(target * tf.log(output + epsilon) + (1. - target) * tf.log(1. - output + epsilon)), axis=1))

    # For brevity, let `x = output`, `z = target`.  The binary cross entropy loss is
    #
    #     loss(x, z) = - sum_i (x[i] * log(z[i]) + (1 - x[i]) * log(1 - z[i]))


def mean_squared_error(output, target, is_mean=False, name="mean_squared_error"):
    """Return the TensorFlow expression of mean-square-error (L2) of two batch of data.

    Parameters
    ----------
    output : Tensor
        2D, 3D or 4D tensor i.e. [batch_size, n_feature], [batch_size, height, width] or [batch_size, height, width, channel].
    target : Tensor
        The target distribution, format the same with `output`.
    is_mean : boolean
        Whether compute the mean or sum for each example.
            - If True, use ``tf.reduce_mean`` to compute the loss between one target and predict data.
            - If False, use ``tf.reduce_sum`` (default).

    References
    ------------
    - `Wiki Mean Squared Error <https://en.wikipedia.org/wiki/Mean_squared_error>`__

    """
    with tf.name_scope(name):
        if output.get_shape().ndims == 2:  # [batch_size, n_feature]
            if is_mean:
                mse = tf.reduce_mean(tf.reduce_mean(tf.squared_difference(output, target), 1))
            else:
                mse = tf.reduce_mean(tf.reduce_sum(tf.squared_difference(output, target), 1))
        elif output.get_shape().ndims == 3:  # [batch_size, w, h]
            if is_mean:
                mse = tf.reduce_mean(tf.reduce_mean(tf.squared_difference(output, target), [1, 2]))
            else:
                mse = tf.reduce_mean(tf.reduce_sum(tf.squared_difference(output, target), [1, 2]))
        elif output.get_shape().ndims == 4:  # [batch_size, w, h, c]
            if is_mean:
                mse = tf.reduce_mean(tf.reduce_mean(tf.squared_difference(output, target), [1, 2, 3]))
            else:
                mse = tf.reduce_mean(tf.reduce_sum(tf.squared_difference(output, target), [1, 2, 3]))
        else:
            raise Exception("Unknow dimension")
        return mse


def normalized_mean_squared_error(output, target):
    """Return the TensorFlow expression of normalized mean-square-error of two distributions.

    Parameters
    ----------
    output : Tensor
        2D, 3D or 4D tensor i.e. [batch_size, n_feature], [batch_size, height, width] or [batch_size, height, width, channel].
    target : Tensor
        The target distribution, format the same with `output`.

    """
    with tf.name_scope("mean_squared_error_loss"):
        if output.get_shape().ndims == 2:  # [batch_size, n_feature]
            nmse_a = tf.sqrt(tf.reduce_sum(tf.squared_difference(output, target), axis=1))
            nmse_b = tf.sqrt(tf.reduce_sum(tf.square(target), axis=1))
        elif output.get_shape().ndims == 3:  # [batch_size, w, h]
            nmse_a = tf.sqrt(tf.reduce_sum(tf.squared_difference(output, target), axis=[1, 2]))
            nmse_b = tf.sqrt(tf.reduce_sum(tf.square(target), axis=[1, 2]))
        elif output.get_shape().ndims == 4:  # [batch_size, w, h, c]
            nmse_a = tf.sqrt(tf.reduce_sum(tf.squared_difference(output, target), axis=[1, 2, 3]))
            nmse_b = tf.sqrt(tf.reduce_sum(tf.square(target), axis=[1, 2, 3]))
        nmse = tf.reduce_mean(nmse_a / nmse_b)
    return nmse


def absolute_difference_error(output, target, is_mean=False):
    """Return the TensorFlow expression of absolute difference error (L1) of two batch of data.

    Parameters
    ----------
    output : Tensor
        2D, 3D or 4D tensor i.e. [batch_size, n_feature], [batch_size, height, width] or [batch_size, height, width, channel].
    target : Tensor
        The target distribution, format the same with `output`.
    is_mean : boolean
        Whether compute the mean or sum for each example.
            - If True, use ``tf.reduce_mean`` to compute the loss between one target and predict data.
            - If False, use ``tf.reduce_sum`` (default).

    """
    with tf.name_scope("mean_squared_error_loss"):
        if output.get_shape().ndims == 2:  # [batch_size, n_feature]
            if is_mean:
                loss = tf.reduce_mean(tf.reduce_mean(tf.abs(output - target), 1))
            else:
                loss = tf.reduce_mean(tf.reduce_sum(tf.abs(output - target), 1))
        elif output.get_shape().ndims == 3:  # [batch_size, w, h]
            if is_mean:
                loss = tf.reduce_mean(tf.reduce_mean(tf.abs(output - target), [1, 2]))
            else:
                loss = tf.reduce_mean(tf.reduce_sum(tf.abs(output - target), [1, 2]))
        elif output.get_shape().ndims == 4:  # [batch_size, w, h, c]
            if is_mean:
                loss = tf.reduce_mean(tf.reduce_mean(tf.abs(output - target), [1, 2, 3]))
            else:
                loss = tf.reduce_mean(tf.reduce_sum(tf.abs(output - target), [1, 2, 3]))
        else:
            raise Exception("Unknow dimension")
        return loss
    

def cosine_distance_error (output, target):
    """Return the TensorFlow expression of cosine distance difference error of two batch of data.
    
    Parameters
    ----------
    output : Tensor
        2D tensor i.e. [batch_size, n_feature]
    target : Tensor
        The target distribution, format the same with `output`.        
    """
    
    return tf.reduce_mean (
        tf.losses.cosine_distance (
            tf.nn.l2_normalize(output, 1), 
            tf.nn.l2_normalize(target, 1),
            1 # dim / axis
        )
    )


def dice_coe(output, target, loss_type='jaccard', axis=(1, 2, 3), smooth=1e-5):
    """Soft dice (Sørensen or Jaccard) coefficient for comparing the similarity
    of two batch of data, usually be used for binary image segmentation
    i.e. labels are binary. The coefficient between 0 to 1, 1 means totally match.

    Parameters
    -----------
    output : Tensor
        A distribution with shape: [batch_size, ....], (any dimensions).
    target : Tensor
        The target distribution, format the same with `output`.
    loss_type : str
        ``jaccard`` or ``sorensen``, default is ``jaccard``.
    axis : tuple of int
        All dimensions are reduced, default ``[1,2,3]``.
    smooth : float
        This small value will be added to the numerator and denominator.
            - If both output and target are empty, it makes sure dice is 1.
            - If either output or target are empty (all pixels are background), dice = ```smooth/(small_value + smooth)``, then if smooth is very small, dice close to 0 (even the image values lower than the threshold), so in this case, higher smooth can have a higher dice.

    Examples
    ---------
    >>> outputs = tl.act.pixel_wise_softmax(network.outputs)
    >>> dice_loss = 1 - tl.cost.dice_coe(outputs, y_)

    References
    -----------
    - `Wiki-Dice <https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient>`__

    """
    inse = tf.reduce_sum(output * target, axis=axis)
    if loss_type == 'jaccard':
        l = tf.reduce_sum(output * output, axis=axis)
        r = tf.reduce_sum(target * target, axis=axis)
    elif loss_type == 'sorensen':
        l = tf.reduce_sum(output, axis=axis)
        r = tf.reduce_sum(target, axis=axis)
    else:
        raise Exception("Unknow loss_type")
    ## old axis=[0,1,2,3]
    # dice = 2 * (inse) / (l + r)
    # epsilon = 1e-5
    # dice = tf.clip_by_value(dice, 0, 1.0-epsilon) # if all empty, dice = 1
    ## new haodong
    dice = (2. * inse + smooth) / (l + r + smooth)
    ##
    dice = tf.reduce_mean(dice)
    return dice


def dice_hard_coe(output, target, threshold=0.5, axis=(1, 2, 3), smooth=1e-5):
    """Non-differentiable Sørensen–Dice coefficient for comparing the similarity
    of two batch of data, usually be used for binary image segmentation i.e. labels are binary.
    The coefficient between 0 to 1, 1 if totally match.

    Parameters
    -----------
    output : tensor
        A distribution with shape: [batch_size, ....], (any dimensions).
    target : tensor
        The target distribution, format the same with `output`.
    threshold : float
        The threshold value to be true.
    axis : tuple of integer
        All dimensions are reduced, default ``(1,2,3)``.
    smooth : float
        This small value will be added to the numerator and denominator, see ``dice_coe``.

    References
    -----------
    - `Wiki-Dice <https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient>`__

    """
    output = tf.cast(output > threshold, dtype=tf.float32)
    target = tf.cast(target > threshold, dtype=tf.float32)
    inse = tf.reduce_sum(tf.multiply(output, target), axis=axis)
    l = tf.reduce_sum(output, axis=axis)
    r = tf.reduce_sum(target, axis=axis)
    ## old axis=[0,1,2,3]
    # hard_dice = 2 * (inse) / (l + r)
    # epsilon = 1e-5
    # hard_dice = tf.clip_by_value(hard_dice, 0, 1.0-epsilon)
    ## new haodong
    hard_dice = (2. * inse + smooth) / (l + r + smooth)
    ##
    hard_dice = tf.reduce_mean(hard_dice)
    return hard_dice


def iou_coe(output, target, threshold=0.5, axis=(1, 2, 3), smooth=1e-5):
    """Non-differentiable Intersection over Union (IoU) for comparing the
    similarity of two batch of data, usually be used for evaluating binary image segmentation.
    The coefficient between 0 to 1, and 1 means totally match.

    Parameters
    -----------
    output : tensor
        A batch of distribution with shape: [batch_size, ....], (any dimensions).
    target : tensor
        The target distribution, format the same with `output`.
    threshold : float
        The threshold value to be true.
    axis : tuple of integer
        All dimensions are reduced, default ``(1,2,3)``.
    smooth : float
        This small value will be added to the numerator and denominator, see ``dice_coe``.

    Notes
    ------
    - IoU cannot be used as training loss, people usually use dice coefficient for training, IoU and hard-dice for evaluating.

    """
    pre = tf.cast(output > threshold, dtype=tf.float32)
    truth = tf.cast(target > threshold, dtype=tf.float32)
    inse = tf.reduce_sum(tf.multiply(pre, truth), axis=axis)  # AND
    union = tf.reduce_sum(tf.cast(tf.add(pre, truth) >= 1, dtype=tf.float32), axis=axis)  # OR
    ## old axis=[0,1,2,3]
    # epsilon = 1e-5
    # batch_iou = inse / (union + epsilon)
    ## new haodong
    batch_iou = (inse + smooth) / (union + smooth)
    iou = tf.reduce_mean(batch_iou)
    return iou  #, pre, truth, inse, union


# ## test soft/hard dice and iou
# import numpy as np
# y = np.zeros((1,10,10,1))
# # y[0,0:5,0:5]=1.0
# o = np.zeros((1,10,10,1))
# # o[:,:,:,:] = 0            # what we want: dice=0   iou=0  OK
# # o[0,0:2,0:2]=0.3          # what we want: dice larger iou=0  OK
# # o[0,0:2,0:2]=0.6          # what we want: dice larger  iou small  OK
# # o[0,0:3,0:3]=0.6          # what we want: dice larger iou larger OK
# # o[0,0:3,0:3]=1            # what we want: dice larger iou same OK
# # o[0,0:5,0:5]=1            # what we want: dice=1 iou=1  OK
# # o[0,0:5,0:5]=0.3          # what we want: dice smaller  iou=0  OK
# # o[0,0:5,0:5]=1e-2           # what we want: dice≈0 iou=0  OK
# # o[0,8:10,8:10]=1.0        # what we want: dice=0 iou=0  OK
# # o[0,8:10,8:10]=1e-10        # what we want: dice=0 iou=0  OK
# # y[:,:,:,:] = o[:,:,:,:] = 0 # what we want: dice=1 iou=1  OK
# ## why in u-net, dice=1 hard-dice=1 iou=1 exist?? print bug?
#
# d = dice_coe(o, y, 'jaccard', smooth=1.)
# hd = dice_hard_coe(o, y, smooth=1e-5)
# i = iou_coe(o, y, smooth=1e-5)
# sess = tf.Session(config=tf.ConfigProto(allow_soft_placement=True))
# # sess.run(tf.local_variables_initializer())
# print(sess.run([d,hd,i]))
# # p, t, i, u = sess.run([pre, truth, inse, union])
# # import pprint
# # pprint.pprint(((y>0.5)*(o>0.5)).astype(int).tolist())
# # pprint.pprint(p.tolist())
# # pprint.pprint(t.tolist())
# # pprint.pprint(i)
# # pprint.pprint(u)
# exit()


def cross_entropy_seq (logits, target_seqs, batch_size=None):  #, batch_size=1, num_steps=None):
    """Returns the expression of cross-entropy of two sequences, implement
    softmax internally. Normally be used for fixed length RNN outputs, see `PTB example <https://github.com/zsdonghao/tensorlayer/blob/master/example/tutorial_ptb_lstm_state_is_tuple.py>`__.

    Parameters
    ----------
    logits : Tensor
        2D tensor with shape of `[batch_size * n_steps, n_classes]`.
    target_seqs : Tensor
        The target sequence, 2D tensor `[batch_size, n_steps]`, if the number of step is dynamic, please use ``tl.cost.cross_entropy_seq_with_mask`` instead.
    batch_size : None or int.
        Whether to divide the cost by batch size.
            - If integer, the return cost will be divided by `batch_size`.
            - If None (default), the return cost will not be divided by anything.

    Examples
    --------
    >>> see `PTB example <https://github.com/zsdonghao/tensorlayer/blob/master/example/tutorial_ptb_lstm_state_is_tuple.py>`__.for more details
    >>> input_data = tf.placeholder(tf.int32, [batch_size, n_steps])
    >>> targets = tf.placeholder(tf.int32, [batch_size, n_steps])
    >>> # build the network
    >>> print(net.outputs)
    ... (batch_size * n_steps, n_classes)
    >>> cost = tl.cost.cross_entropy_seq(network.outputs, targets)

    """
    # try: # TF 1.0
    sequence_loss_by_example_fn = tf.contrib.legacy_seq2seq.sequence_loss_by_example
    # except:
    #     sequence_loss_by_example_fn = tf.nn.seq2seq.sequence_loss_by_example

    loss = sequence_loss_by_example_fn([logits], [tf.reshape(target_seqs, [-1])], [tf.ones_like(tf.reshape(target_seqs, [-1]), dtype=tf.float32)])
    # [tf.ones([batch_size * num_steps])])
    cost = tf.reduce_sum(loss)  #/ batch_size
    if batch_size is not None:
        cost = cost / batch_size
    return cost

def mean_seq (logits, target_seqs, batch_size, sequence_length):
    weights = tf.ones ([batch_size, sequence_length])
    sequence_loss = tf.contrib.seq2seq.sequence_loss (
        logits = logits, targets = target_seqs, weights = weights
    )
    return tf.reduce_mean (sequence_loss)

def cross_entropy_seq_with_mask(logits, target_seqs, input_mask, return_details=False, name=None):
    """Returns the expression of cross-entropy of two sequences, implement
    softmax internally. Normally be used for Dynamic RNN with Synced sequence input and output.

    Parameters
    -----------
    logits : Tensor
        2D tensor with shape of [batch_size * ?, n_classes], `?` means dynamic IDs for each example.
        - Can be get from `DynamicRNNLayer` by setting ``return_seq_2d`` to `True`.
    target_seqs : Tensor
        int of tensor, like word ID. [batch_size, ?], `?` means dynamic IDs for each example.
    input_mask : Tensor
        The mask to compute loss, it has the same size with `target_seqs`, normally 0 or 1.
    return_details : boolean
        Whether to return detailed losses.
            - If False (default), only returns the loss.
            - If True, returns the loss, losses, weights and targets (see source code).

    Examples
    --------
    >>> batch_size = 64
    >>> vocab_size = 10000
    >>> embedding_size = 256
    >>> input_seqs = tf.placeholder(dtype=tf.int64, shape=[batch_size, None], name="input")
    >>> target_seqs = tf.placeholder(dtype=tf.int64, shape=[batch_size, None], name="target")
    >>> input_mask = tf.placeholder(dtype=tf.int64, shape=[batch_size, None], name="mask")
    >>> net = tl.layers.EmbeddingInputlayer(
    ...         inputs = input_seqs,
    ...         vocabulary_size = vocab_size,
    ...         embedding_size = embedding_size,
    ...         name = 'seq_embedding')
    >>> net = tl.layers.DynamicRNNLayer(net,
    ...         cell_fn = tf.contrib.rnn.BasicLSTMCell,
    ...         n_hidden = embedding_size,
    ...         dropout = (0.7 if is_train else None),
    ...         sequence_length = tl.layers.retrieve_seq_length_op2(input_seqs),
    ...         return_seq_2d = True,
    ...         name = 'dynamicrnn')
    >>> print(net.outputs)
    ... (?, 256)
    >>> net = tl.layers.DenseLayer(net, n_units=vocab_size, name="output")
    >>> print(net.outputs)
    ... (?, 10000)
    >>> loss = tl.cost.cross_entropy_seq_with_mask(net.outputs, target_seqs, input_mask)

    """
    targets = tf.reshape(target_seqs, [-1])  # to one vector
    weights = tf.to_float(tf.reshape(input_mask, [-1]))  # to one vector like targets
    losses = tf.nn.sparse_softmax_cross_entropy_with_logits(logits=logits, labels=targets, name=name) * weights
    #losses = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(logits=logits, labels=targets, name=name)) # for TF1.0 and others

    # try: ## TF1.0
    loss = tf.divide(
        tf.reduce_sum(losses),  # loss from mask. reduce_sum before element-wise mul with mask !!
        tf.reduce_sum(weights),
        name="seq_loss_with_mask")
    # except: ## TF0.12
    #     loss = tf.div(tf.reduce_sum(losses),   # loss from mask. reduce_sum before element-wise mul with mask !!
    #                     tf.reduce_sum(weights),
    #                     name="seq_loss_with_mask")
    if return_details:
        return loss, losses, weights, targets
    else:
        return loss


def cosine_similarity(v1, v2):
    """Cosine similarity [-1, 1].

    Parameters
    ----------
    v1, v2 : Tensor
        Tensor with the same shape [batch_size, n_feature].

    Returns
    -------
    Tensor
        a tensor of shape [batch_size].

    References
    ----------
    - `<https://en.wikipedia.org/wiki/Cosine_similarity>`__.

    """
    # try: ## TF1.0
    cost = tf.reduce_sum(tf.multiply(v1, v2), 1) / (tf.sqrt(tf.reduce_sum(tf.multiply(v1, v1), 1)) * tf.sqrt(tf.reduce_sum(tf.multiply(v2, v2), 1)))
    # except: ## TF0.12
    #     cost = tf.reduce_sum(tf.mul(v1, v2), reduction_indices=1) / (tf.sqrt(tf.reduce_sum(tf.mul(v1, v1), reduction_indices=1)) * tf.sqrt(tf.reduce_sum(tf.mul(v2, v2), reduction_indices=1)))
    return cost


## Regularization Functions
def li_regularizer(scale, scope=None):
    """Li regularization removes the neurons of previous layer. The `i` represents `inputs`.
    Returns a function that can be used to apply group li regularization to weights.
    The implementation follows `TensorFlow contrib <https://github.com/tensorflow/tensorflow/blob/master/tensorflow/contrib/layers/python/layers/regularizers.py>`__.

    Parameters
    ----------
    scale : float
        A scalar multiplier `Tensor`. 0.0 disables the regularizer.
    scope: str
        An optional scope name for this function.

    Returns
    --------
    A function with signature `li(weights, name=None)` that apply Li regularization.

    Raises
    ------
    ValueError : if scale is outside of the range [0.0, 1.0] or if scale is not a float.

    """
    import numbers
    from tensorflow.python.framework import ops
    from tensorflow.python.ops import standard_ops
    # from tensorflow.python.platform import tf_logging as logging

    if isinstance(scale, numbers.Integral):
        raise ValueError('scale cannot be an integer: %s' % scale)
    if isinstance(scale, numbers.Real):
        if scale < 0.:
            raise ValueError('Setting a scale less than 0 on a regularizer: %g' % scale)
        if scale >= 1.:
            raise ValueError('Setting a scale greater than 1 on a regularizer: %g' % scale)
        if scale == 0.:
            logging.info('Scale of 0 disables regularizer.')
            return lambda _, name=None: None

    def li(weights):
        """Applies li regularization to weights."""
        with tf.name_scope('li_regularizer') as scope:
            my_scale = ops.convert_to_tensor(scale, dtype=weights.dtype.base_dtype, name='scale')
            # if tf.__version__ <= '0.12':
            #     standard_ops_fn = standard_ops.mul
            # else:
            standard_ops_fn = standard_ops.multiply
            return standard_ops_fn(my_scale, standard_ops.reduce_sum(standard_ops.sqrt(standard_ops.reduce_sum(tf.square(weights), 1))), name=scope)

    return li


def lo_regularizer(scale):
    """Lo regularization removes the neurons of current layer. The `o` represents `outputs`
    Returns a function that can be used to apply group lo regularization to weights.
    The implementation follows `TensorFlow contrib <https://github.com/tensorflow/tensorflow/blob/master/tensorflow/contrib/layers/python/layers/regularizers.py>`__.

    Parameters
    ----------
    scale : float
        A scalar multiplier `Tensor`. 0.0 disables the regularizer.

    Returns
    -------
    A function with signature `lo(weights, name=None)` that apply Lo regularization.

    Raises
    ------
    ValueError : If scale is outside of the range [0.0, 1.0] or if scale is not a float.

    """
    import numbers
    from tensorflow.python.framework import ops
    from tensorflow.python.ops import standard_ops
    # from tensorflow.python.platform import tf_logging as logging

    if isinstance(scale, numbers.Integral):
        raise ValueError('scale cannot be an integer: %s' % scale)
    if isinstance(scale, numbers.Real):
        if scale < 0.:
            raise ValueError('Setting a scale less than 0 on a regularizer: %g' % scale)
        if scale >= 1.:
            raise ValueError('Setting a scale greater than 1 on a regularizer: %g' % scale)
        if scale == 0.:
            logging.info('Scale of 0 disables regularizer.')
            return lambda _, name=None: None

    def lo(weights, name='lo_regularizer'):
        """Applies group column regularization to weights."""
        with tf.name_scope(name) as scope:
            my_scale = ops.convert_to_tensor(scale, dtype=weights.dtype.base_dtype, name='scale')
            # if tf.__version__ <= '0.12':
            #     standard_ops_fn = standard_ops.mul
            # else:
            standard_ops_fn = standard_ops.multiply
            return standard_ops_fn(my_scale, standard_ops.reduce_sum(standard_ops.sqrt(standard_ops.reduce_sum(tf.square(weights), 0))), name=scope)

    return lo


def maxnorm_regularizer(scale=1.0):
    """Max-norm regularization returns a function that can be used to apply max-norm regularization to weights.

    More about max-norm, see `wiki-max norm <https://en.wikipedia.org/wiki/Matrix_norm#Max_norm>`_.
    The implementation follows `TensorFlow contrib <https://github.com/tensorflow/tensorflow/blob/master/tensorflow/contrib/layers/python/layers/regularizers.py>`__.

    Parameters
    ----------
    scale : float
        A scalar multiplier `Tensor`. 0.0 disables the regularizer.

    Returns
    ---------
    A function with signature `mn(weights, name=None)` that apply Lo regularization.

    Raises
    --------
    ValueError : If scale is outside of the range [0.0, 1.0] or if scale is not a float.

    """
    import numbers
    from tensorflow.python.framework import ops
    from tensorflow.python.ops import standard_ops

    if isinstance(scale, numbers.Integral):
        raise ValueError('scale cannot be an integer: %s' % scale)
    if isinstance(scale, numbers.Real):
        if scale < 0.:
            raise ValueError('Setting a scale less than 0 on a regularizer: %g' % scale)
        # if scale >= 1.:
        #   raise ValueError('Setting a scale greater than 1 on a regularizer: %g' %
        #                    scale)
        if scale == 0.:
            logging.info('Scale of 0 disables regularizer.')
            return lambda _, name=None: None

    def mn(weights, name='max_regularizer'):
        """Applies max-norm regularization to weights."""
        with tf.name_scope(name) as scope:
            my_scale = ops.convert_to_tensor(scale, dtype=weights.dtype.base_dtype, name='scale')
            #   if tf.__version__ <= '0.12':
            #       standard_ops_fn = standard_ops.mul
            #   else:
            standard_ops_fn = standard_ops.multiply
            return standard_ops_fn(my_scale, standard_ops.reduce_max(standard_ops.abs(weights)), name=scope)

    return mn


def maxnorm_o_regularizer(scale):
    """Max-norm output regularization removes the neurons of current layer.
    Returns a function that can be used to apply max-norm regularization to each column of weight matrix.
    The implementation follows `TensorFlow contrib <https://github.com/tensorflow/tensorflow/blob/master/tensorflow/contrib/layers/python/layers/regularizers.py>`__.

    Parameters
    ----------
    scale : float
        A scalar multiplier `Tensor`. 0.0 disables the regularizer.

    Returns
    ---------
    A function with signature `mn_o(weights, name=None)` that apply Lo regularization.

    Raises
    ---------
    ValueError : If scale is outside of the range [0.0, 1.0] or if scale is not a float.

    """
    import numbers
    from tensorflow.python.framework import ops
    from tensorflow.python.ops import standard_ops

    if isinstance(scale, numbers.Integral):
        raise ValueError('scale cannot be an integer: %s' % scale)
    if isinstance(scale, numbers.Real):
        if scale < 0.:
            raise ValueError('Setting a scale less than 0 on a regularizer: %g' % scale)
        # if scale >= 1.:
        #   raise ValueError('Setting a scale greater than 1 on a regularizer: %g' %
        #                    scale)
        if scale == 0.:
            logging.info('Scale of 0 disables regularizer.')
            return lambda _, name=None: None

    def mn_o(weights, name='maxnorm_o_regularizer'):
        """Applies max-norm regularization to weights."""
        with tf.name_scope(name) as scope:
            my_scale = ops.convert_to_tensor(scale, dtype=weights.dtype.base_dtype, name='scale')
            if tf.__version__ <= '0.12':
                standard_ops_fn = standard_ops.mul
            else:
                standard_ops_fn = standard_ops.multiply
            return standard_ops_fn(my_scale, standard_ops.reduce_sum(standard_ops.reduce_max(standard_ops.abs(weights), 0)), name=scope)

    return mn_o


def maxnorm_i_regularizer(scale):
    """Max-norm input regularization removes the neurons of previous layer.
    Returns a function that can be used to apply max-norm regularization to each row of weight matrix.
    The implementation follows `TensorFlow contrib <https://github.com/tensorflow/tensorflow/blob/master/tensorflow/contrib/layers/python/layers/regularizers.py>`__.

    Parameters
    ----------
    scale : float
        A scalar multiplier `Tensor`. 0.0 disables the regularizer.

    Returns
    ---------
    A function with signature `mn_i(weights, name=None)` that apply Lo regularization.

    Raises
    ---------
    ValueError : If scale is outside of the range [0.0, 1.0] or if scale is not a float.

    """
    import numbers
    from tensorflow.python.framework import ops
    from tensorflow.python.ops import standard_ops

    if isinstance(scale, numbers.Integral):
        raise ValueError('scale cannot be an integer: %s' % scale)
    if isinstance(scale, numbers.Real):
        if scale < 0.:
            raise ValueError('Setting a scale less than 0 on a regularizer: %g' % scale)
        # if scale >= 1.:
        #   raise ValueError('Setting a scale greater than 1 on a regularizer: %g' %
        #                    scale)
        if scale == 0.:
            logging.info('Scale of 0 disables regularizer.')
            return lambda _, name=None: None

    def mn_i(weights, name='maxnorm_i_regularizer'):
        """Applies max-norm regularization to weights."""
        with tf.name_scope(name) as scope:
            my_scale = ops.convert_to_tensor(scale, dtype=weights.dtype.base_dtype, name='scale')
            if tf.__version__ <= '0.12':
                standard_ops_fn = standard_ops.mul
            else:
                standard_ops_fn = standard_ops.multiply
            return standard_ops_fn(my_scale, standard_ops.reduce_sum(standard_ops.reduce_max(standard_ops.abs(weights), 1)), name=scope)

    return mn_i


import tensorflow as tf
import numpy as np
import sys
import os, shutil
import random
from rs4 import pathtool
from . import overfit

class Task:
    def __init__ (self, dnns, attr):
        self.dnns = dnns
        self.attr = attr
        
    def __call__ (self, *args, **karg):
        results = []
        for dnn in self.dnns:
            results.append (getattr (dnn, self.attr) (*args, **karg))
        return np.array (results).transpose ()


class MultiDNN:
    def __init__ (self, *args):
        self.dnns = []
        self.y_div = [0]
        for i, arg in enumerate (args):
            if i % 2 == 0:
                self.dnns.append (arg)
                assert arg.name is not None
            else:
                self.y_div.append (sum (self.y_div) + arg)
        self.verbose_turned_offs = []
                        
    def __getattr__ (self, attr):
        return Task (self.dnns, attr)  
    
    def _get_segment (self, i, ys):
        return ys [:,self.y_div [i]:self.y_div [i+1]]
    
    def normalize (self, *args, **karg):
        xs = []
        for i, dnn in enumerate (self.dnns):
            xs = dnn.normalize (*args, **karg)
        return xs

    def reset_dir (self, target):
        self.dnns [0].reset_dir (target)
    
    def reset_tensor_board (self, summaries_dir):
        for i, dnn in enumerate (self.dnns):
            dnn.reset_tensor_board (summaries_dir, i == 0)
        
    def is_overfit (self):
        results = []
        for i, dnn in enumerate (self.dnns):
            results.append (int (dnn.is_overfit ()))
        return sum (results) == len (self.dnns)
            
    def write_summary (self, writer, feed_dict):
        for i in range (len (list (feed_dict.values ())[0])):
            feed_dict_ = {}
            for k, v in feed_dict.items ():
                feed_dict_ [k] = v [i]
            dnn = self.dnns [i]
            dnn.write_summary (writer, feed_dict_)
                    
    def custom_accuracy (self, preds, ys, *args, **karg):
        results = []
        for i in range (len (self.y_div) - 1):
            dnn = self.dnns [i]
            results.append (dnn.custom_accuracy (preds [i], self._get_segment (i, ys), *args, **karg))
        return np.array (results)
    
    def __split_run (self, attr, x, y, **kargs):
        class Result:
            def __init__ (self, results):
                self.results = results
            
            def update (self, *args, **karg): 
                for r in self.results:                    
                    r.update (*args, **karg)
            
            def __getattr__ (self, attr):
                return np.array ([getattr (r, attr) for r in self.results])
                    
        ys = y
        results = []
        for i in range (len (self.y_div) - 1):
            dnn = self.dnns [i]
            kargs ["x"] = x
            kargs ["y"] = self._get_segment (i, ys)
            result = getattr (dnn, attr) (**kargs)
            results.append (result)
        if attr == "fit":
            return [r [0] for r in results] 
        elif attr in ("train", "valid"):
            return Result (results)
        return np.array (results, np.object).transpose ()
    
    def batch (self, x, y, **kargs):
        return self.__split_run ("batch", x, y, **kargs)
        
    def resub (self, x, y, **kargs):
        return self.__split_run ("resub", x, y, **kargs)
        
    def valid (self, x, y, **kargs):
        return self.__split_run ("valid", x, y, **kargs)
    
    def run (self, *ops, **kargs):
        ys = kargs.pop ("y")
        results = []
        for i in range (len (self.y_div) - 1):
            dnn = self.dnns [i]
            kargs ["y"] = self._get_segment (i, ys)
            result = list (dnn.run (*tuple ([getattr (dnn, op.attr) for op in ops]), **kargs))
            results.append (result)                    
        return np.array (results, np.object).transpose ()
        


import re
import time

currnet_year = str (time.localtime (time.time ()) [0])
mons = ['jan','feb','mar','apr','may','jun','jul','aug','sep','oct','nov','dec']
monf = ['january','february','march','april','may','june','july','august','september','october','november','december']

monalpha = "(?P<month>%s)\.?" % "|".join (monf + mons)
mondigit = '(?P<month>0?[1-9]|1[0-2])'
fmondigit = '(?P<month>0[1-9]|1[0-2])'
day = '(?P<day>0?[1-9]|[12][0-9]|3[01])(?:st|nd|rd|th)?(?:\sday)?(?:\sof)?'
fday = '(?P<day>0[1-9]|[12][0-9]|3[01])(?:st|nd|rd|th)?'
year4 = '(?P<year>20[01234][0-9])'
year2 = '(?P<year>[1234][0-9])'
delima = '([-\s,]+)'
delims = '([-./]+)'
digit_nofollow = '(?![0-9])'
digit_nolead = '(?<![0-9])'
date_formats = ('mdy', 'ymd', 'dmy')

MDY = (
    '[:monalpha:][:delima:][:day:][:delima:][:year4:][:dnf:]',
    '[:monalpha:][:delima:][:day:][:delima:][:year2:][:dnf:]',
    '[:dnl:][:mondigit:][:delims:][:day:][:delims:][:year4:][:dnf:]',
    '[:dnl:][:mondigit:][:delims:][:day:][:delims:][:year2:][:dnf:]',
    '[:monalpha:][:delima:][:day:][:dnf:]',
    '[:dnl:][:fmondigit:][:fday:][:year4:][:dnf:]',
    '[:dnl:][:fmondigit:][:fday:][:year2:][:dnf:]',
)

DMY = (    
    '[:dnl:][:day:][:delima:][:monalpha:][:delima:][:year4:][:dnf:]',
    '[:dnl:][:day:][:delima:][:monalpha:][:delima:][:year2:][:dnf:]',
    '[:dnl:][:day:][:delims:][:mondigit:][:delims:][:year4:][:dnf:]',
    '[:dnl:][:day:][:delims:][:mondigit:][:delims:][:year2:][:dnf:]',
    '[:dnl:][:day:][:delima:][:monalpha:]',    
    '[:dnl:][:fday:][:fmondigit:][:year4:][:dnf:]',
    '[:dnl:][:fday:][:fmondigit:][:year2:][:dnf:]',
)

YMD = (    
    '[:dnl:][:year4:][:delima:][:monalpha:][:delima:][:day:][:dnf:]',
    '[:dnl:][:year2:][:delima:][:monalpha:][:delima:][:day:][:dnf:]',
    '[:dnl:][:year4:][:delims:][:mondigit:][:delims:][:day:][:dnf:]',
    '[:dnl:][:year2:][:delims:][:mondigit:][:delims:][:day:][:dnf:]',    
    '[:dnl:][:year4:][:fmondigit:][:fday:][:dnf:]',
    '[:dnl:][:year2:][:fmondigit:][:fday:][:dnf:]',
)

FORMATS = {}
for fmt, group in (('mdy', MDY), ('dmy', DMY), ('ymd', YMD)):
    ngroup = []
    for each in group:
        each = each.replace ('[:monalpha:]', monalpha)
        each = each.replace ('[:mondigit:]', mondigit)
        each = each.replace ('[:fmondigit:]', fmondigit)
        each = each.replace ('[:delims:]', delims)
        each = each.replace ('[:delima:]', delima)
        each = each.replace ('[:day:]', day)
        each = each.replace ('[:fday:]',fday)
        each = each.replace ('[:year4:]', year4)
        each = each.replace ('[:year2:]', year2)
        each = each.replace ('[:dnf:]', digit_nofollow)
        each = each.replace ('[:dnl:]', digit_nolead)        
        ngroup.append (each)
    FORMATS [fmt] = ngroup

RXS = []
for dateformat in date_formats:
    for b in FORMATS [dateformat]:            
        exp = "(%s)" % (b)
        RXS.append (re.compile (exp, re.S))
        
#-------------------------------------------------------------
def normtext (content):
    rt = content.lower()
    rt = re.sub ('</?.*?>',' ',rt)        
    rt = rt.replace('\n',' ')
    rt = rt.replace('\r',' ')
    rt = rt.replace('\t',' ')
    rt = re.sub ('[ ]{2,}',' ', rt)    
    return rt

#-------------------------------------------------------------
def replace (content, token = '<date>'):
    content = normtext (content)
    for rx in RXS:
       content = rx.sub (token, content)            
    return content.strip ()   


if __name__ == "__main__":    
    print (parse_date ("""
<p>Two Rivers School District
Notice of invitation for bid (IFB)
The Two Rivers District will be accepting bids for the foodservice dept. for the 2017-18 school year, including grocery's, milk and juice, bread, and supply's. Bid packets may be picked up at the Superintendent's office during the week of June 5th-23rd. 
All bids must be received by Thursday, July 6th at 10:00am. Bids should be in a sealed envelope and will be opened Friday July 7th at 9:00 am. The Two Rivers School District follows all rules pertaining to Debarment and Suspension and will use small, minority, and women's businesses whenever possible. For further questions please call: Becky Fowler, Food Service Director at 479-272-3113.
74157186f 
    """, "mdy")
    )
    
    #print (parse_date ("jun 1, 17", "mdy", 0))
    #print (parse_date ("Closes 05/30/2017 05:00 PM CT Solicitation No 750.5-17-01 Title Off-Highway Vehicle (OHV) Use Map for the Little Missouri National Grasslands Issuing Agency Parks & Recreation, Department of", "mdy", 1))
    
    print (parse_date ('June 18, 2018 at 2 pm', "mdy", 0))
    print (parse_date ('RFQ packets must be returned by 12 p.m. on Wednesday, May 1, 2019.', "mdy", 1))
    


import numpy as np
import math
from sklearn.utils.extmath import softmax as softmax_
import hyperopt
import os
from rs4 import siesta

def softmax (x):
    x = np.array (x)
    if len (x.shape) == 2:
        return softmax_ (x).tolist ()
    return softmax_ ([x])[0].tolist ()

def sigmoid (x):
    return [1 / (1 + np.exp(-e)) for e in x]

def confusion_matrix (labels, predictions, num_labels):
  rows = []
  for i in range (num_labels):
    row = np.bincount (predictions[labels == i], minlength=num_labels)
    rows.append (row)
  return np.vstack (rows)

def render_trial (space, trial, stringfy = False):
    if trial.get ("misc"):
        trial = dict ([(k, v [0]) for k, v in trial ["misc"]["vals"].items ()])
    params = hyperopt.space_eval(space, trial)
    if stringfy:
        return ", ".join (["{}: {}".format (k, v) for k, v in params.items ()])
    return params

def get_latest_model (path):
    if not os.path.isdir (path) or not os.listdir (path):
        return
    version = max ([int (ver) for ver in os.listdir (path) if ver.isdigit () and os.path.isdir (os.path.join (path, ver))])    
    return os.path.join (path, str (version))
 
 
# Model Loaders ------------------------------------------------

class CheckPoint:
    def __init__ (self, net):
        self.net = net
    
    def ftest (self, xs, ys, **kargs):
        return self.net.ftest (self.net.normalize (xs), ys, **kargs)
            
    def predict (self, xs):
        return self.net.run (self.net.logit, x = self.net.normalize (xs)) [0]
        

class SavedModel:
    def __init__ (self, model_root, debug = False):
        self.model_root = model_root
        self.debug = debug
        self.model_path = get_latest_model (self.model_root)        
        self.stub = self.create_stub (self.model_path)
        self.version = os.path.basename (self.model_path)        
    
    def create_stub (self, model_path):
        from . import saved_model
        return saved_model.load (model_path)
                
    def predict (self, x, **kargs):
        y = self.stub.run (x, **kargs)[0]
        return y.astype (np.float64)

    
class TFLite (SavedModel):    
    def __init__ (self, model_root, model_file = "model.tflite", debug = False):
        self.model_file = model_file
        SavedModel.__init__ (self, model_root, debug)
    
    def create_stub (self, model_path):
        from . import tflite
        
        return tflite.load (
            os.path.join (model_path, self.model_file),
            (128, 128)
        )
    
    def predict (self, x, **kargs):   
        ys = []
        for x_ in x:     
            ys.append (self.stub.run (x_, **kargs)[0].astype (np.float64))
        return np.array (ys)    


class TFServer:
    def __init__ (self, endpoint, alias, debug = False):
        from tfserver import cli
                
        self.endpoint = endpoint
        self.alias = alias
        self.debug = debug
        
        self.stub = cli.Server (endpoint)
        api = siesta.API (endpoint)
        resp = api.model (self.alias).version.get ()
        self.version = resp.data.version         
        
    def predict (self, x, **kargs):
        resp = self.stub.predict (self.alias, 'predict', x = x, **kargs)
        return resp.y
            



import tensorflow as tf
import numpy as np
from . import predutil
from rs4.termcolor import tc
from sklearn.utils.extmath import softmax
import math


def confidence_interval (confidence_level, error, n):
    const = {90: 1.64, 95: 1.96, 96: 2.05, 97: 2.17, 98: 2.33, 99: 2.58} [confidence_level]
    return const * math.sqrt ((error * (1 - error)) / n)

class Result:
    train_dir = None
    labels = None
    name = None
    cost_names = []
    _perfmetric_func = None
    _summary_func = None        
            
    def __init__ (self, xs, ys, logits, cost, ops, args, epoch, global_step, learning_rate, is_validating, is_evaluating, on_ftest):
        self.xs = xs
        self.ys = ys
        self.logits = logits
        self.cost = cost        
        self.epoch = epoch
        self.global_step = global_step
        self.learning_rate = learning_rate
        self.is_validating = is_validating
        self.is_evaluating = is_evaluating
        self.ops = ops
        self.args = args
        
        # set by trainer
        self.is_cherry = False
        self.on_ftest = on_ftest
        self.is_overfit = False
        self.is_improved = False
        self.performance = []
        self.performance_names = []
        self.measure_performance ()
    
    def measure_performance (self):    
        if not (self.is_evaluating and self._perfmetric_func): # no run within batch steps
            return  
          
        try:
            performance = self._perfmetric_func (self)
        except NotImplementedError:
            return        
        
        if not isinstance (performance, (tuple, list)):
            performance = [performance]
        
        for idx, p in enumerate (performance):
            if isinstance (p, (list, tuple)):
                self.performance_names.append (p [0])
                self.performance.append (p [1])
            else:
                try:
                    name = "{}-{}".format (idx + 1, self.labels [idx].name)
                except:
                    name = str (idx + 1)    
                self.performance_names.append (name)
                self.performance.append (p)        
                            
    @property
    def x (self):
        return self.xs
    
    @property
    def y (self):
        return self.ys
    
    @property
    def logit (self):
        return self.logits
    
    @property
    def phase (self):
        if not self.is_evaluating:
            return "batch"
        elif self.on_ftest:
            return "final"
        elif self.is_cherry:
            return "saved"            
        elif self.is_validating == 2:
            return "ptest"
        return self.is_validating and 'valid' or 'resub'
        
    def __make_summary (self, d = {}):
        if len (self.cost) == 1:
            d  ["eval/cost"] = self.cost [0]
        else:    
            for idx, cost in enumerate (self.cost):
                try: 
                    name = self.cost_names [idx]
                except IndexError:
                    name = str (idx + 1)
                d  ["eval/cost/{}".format (name)] = cost
        
        if len (self.performance) == 1:        
            d ["eval/acc"] = self.performance [0]
        else:      
            for idx, val in enumerate (self.performance):
                name = "eval/acc/" + self.performance_names [idx]
                d [name] = val
        
        d_ = {}
        for k, v in d.items ():
            if self.name:
                k = "{}:{}".format (self.name, k)
            if isinstance (v, (list, tuple)):
                if len (v) == 1: v = v [0]
                else: raise ValueError ("Required float, int or an array contains only one of them")
            d_ [k] = v
        return d_
    
    def __get_phase_summary (self, kargs):
        output = []
        for k, v in self.__make_summary (kargs).items ():                            
            k = k [5:] # rempve eval/
            if isinstance (v, (float, np.float64, np.float32)):
                output.append ("{} {:.5f}".format (k, v))
            elif isinstance (v, (int, np.int64, np.int32)):
                output.append ("{} {:04d}".format (k, v))
            else:
                raise ValueError ("Required float, int type")
        output.sort ()
        if self.learning_rate is not None:
            output.append ("lr {:.2e}".format (self.learning_rate))

        if self.on_ftest:
            return " | ".join (output)            
        if self.phase != "saved":
            if self.is_overfit:
                output.append (tc.error ("overfitted"))
            if self.is_improved:
                output.append (tc.info ("improved"))
        return " | ".join (output)
    
    def summary (self):
        self._summary_func (self.phase, self.__make_summary (self.args.get ("summary", {})))
        return self
    
    # statistics helper methods -----------------------------------------    
    def log (self, msg = None, **kargs):
        coloring = False
        if not msg:
            msg = self.__get_phase_summary (kargs)
            coloring = True
        phase = self.phase
        if phase == "final":
            header = "[fin.:ftest]"
            color = tc.fail
        else:    
            header = "[{:04d}:{}]".format (self.epoch, self.phase)                                  
            if phase == "saved":
                color = tc.warn
            elif phase == "ptest":
                color = tc.primary   
            else:
                color = tc.debug                   
        print ("{} {}".format ((coloring and color or tc.critical) (header), msg))
        return self

    def reunion (self, data_length, unify_func = np.mean, softmaxing = True):
        logits = []
        ys = []
        index = 0
        for part_length in data_length:        
            ys.append (self.ys [index])
            aud = self.logits [index: index + part_length]
            if softmaxing:
                aud = softmax (aud)                
            index += part_length
            logits.append (unify_func (aud, 0))
        self.logits = np.array (logits)
        self.ys = np.array (ys)
    
    def get_performance (self, index = "avg"):
        ps = self.performance
        if index == "avg":
            return np.mean (ps)
        elif index == "sum":
            return np.sum (ps)
        return ps [index] 
            
    def get_confusion_matrix (self, logits = None, ys = None):
        if logits is None:
            logits = self.logits
        if ys is None:
            ys = self.ys
        mat_ = predutil.confusion_matrix (
            np.argmax (logits, 1), 
            np.argmax (ys, 1),
            logits.shape [1]
        )
        return mat_.T
    
    def slice_by_label (self, label_index):
        start_index = 0
        for i in range (label_index + 1):
            num_label = len (self.labels [i])
            if label_index == i:
                break
            start_index += num_label        
        logits = self.logits [:, start_index:start_index + num_label]
        if len (self.y.shape) == 1: # sparse categorical encoding           
            ys = np.zeros ([len (self.ys), len (self.labels [label_index])])
            for row, y in enumerate (self.ys):
                ys [row][y] = 1.0
        else:        
            ys = self.ys [:, start_index:start_index + num_label]
            if len (ys [0]) == 0:
                ys = self.ys [:, -num_label:]
        return logits, ys
    
    @property    
    def metric (self):
        class Metric:
            def __init__ (self):
                class Tri:
                    def __init__ (self):
                        self.micro = []
                        self.macro = []
                        self.weighted = []                        
                self.accuracy = []
                self.precision = Tri ()
                self.recall = Tri ()
                self.f1 = Tri ()
            
        metric = Metric ()
        if self.labels is None:
            self.calculate_metric (metric, self.logits, self.ys)
        else:            
            for idx, label in enumerate (self.labels):
                logits, ys = self.slice_by_label (idx)
                self.calculate_metric (metric, logits, ys)
        return metric
    
    def calculate_metric (self, metric, logits, ys):
        mat = self.get_confusion_matrix (logits, ys)
        catpreds = mat.sum (axis = 0)
        table = np.zeros ([mat.shape [0], 3])
        for idx, preds in enumerate (mat.T):
            tp = preds [idx]
            fp = catpreds [idx] - tp            
            table [idx][0] = tp
            table [idx][1] = fp
            
        catans = mat.sum (axis = 1)
        for idx, ans in enumerate (mat):
            tp = ans [idx]
            fn = catans [idx] - tp
            table [idx][2] = fn                
        
        accuracy = np.sum (table [:, 0]) / len (logits)
        micro_precision = np.sum (table [:, :1]) / np.sum (table [:, :2])
        micro_recall = np.sum (table [:, :1]) / np.sum ([table [:, 0], table [:, 2]])
        metric.accuracy.append (accuracy)                        
        metric.precision.micro.append (micro_precision)
        metric.recall.micro.append (micro_recall)
        metric.f1.micro.append (2 / (1 / micro_precision + 1 / micro_recall))
        
        precisions = []
        recalls = []            
        f1s = []
        alpha = 1e-8
        for cat, (tp, fp, fn) in enumerate (table):
            precision = tp / (tp + fp + alpha)                
            recall = tp / (tp + fn + alpha)
            f1 = (2 * (precision * recall)) / (precision + recall + alpha)
            precisions.append (precision)
            recalls.append (precision)
            f1s.append (f1)
        
        metric.precision.macro.append (np.mean (precisions))
        metric.precision.weighted.append (np.average (precisions, weights = catans))
        metric.recall.macro.append (np.mean (recalls))                    
        metric.recall.weighted.append (np.average (recalls, weights = catans))
        metric.f1.macro.append (np.mean (f1s))
        metric.f1.weighted.append (np.average (f1s, weights = catans))
            
    def confusion_matrix (self, num_label = 0, label_index = 0, indent = 13, show_label = True):
        start_index = 0
        name = None
        if num_label == 0:
            if not self.labels:
                raise ValueError ("num_label required")            
            logits, ys = self.slice_by_label (label_index)
            name = self.labels [label_index].name
        else:
            logits = self.logits [:, :num_label]
            ys = self.ys [:, :num_label]
            
        mat_ = self.get_confusion_matrix (logits, ys)        
        mat = str (mat_) [1:-1]
        self.log ("confusion matrix{}".format (name and (" of " + name) or ""))
        
        labels = []
        if show_label and self.labels:
            cur_label = self.labels [label_index]
            first_row_length = len (mat.split ("\n", 1) [0]) - 2
            label_width = (first_row_length - 1) // mat_.shape [-1]
            labels = [str (each) [:label_width].rjust (label_width) for each in cur_label.items ()]            
            print (tc.fail ((" " * (indent + label_width + 1)) + " ".join (labels)))
            
        lines = []
        for idx, line in enumerate (mat.split ("\n")):
            if idx > 0:
                line = line [1:]
            line = line [1:-1]                    
            if labels:    
                line = tc.info (labels [idx]) + " " + line
            if indent:    
                line = (" " * indent) + line
            print (line)

    def logit_range (self):
        output_range = [self.logit.min (), self.logit.max (), np.mean (self.logit), np.std (self.logit)]
        quant_range = {}
        for idx, m in enumerate (self.logit [:,:29].argsort (1)[:,-2]):
            sec = int (self.logit [idx, m])
            try: quant_range [sec] += 1
            except KeyError: quant_range [sec] = 1
        quant_range = quant_range
        if quant_range:           
            stats = sorted (quant_range.items ())
            # output range for top1: {} ~ {}, logit range: {:.3f} ~ {:.3f}, mean: {:.3f} std: {:.3f}
            return stats [0][0] - 1, stats [-1][0] + 1, output_range [0], output_range [1], output_range [2], output_range [3]                                
        
    def confidence_level (self, label_index = 0):
        label = self.labels [label_index]
        stat = {}
        for idx in range (len (self.y)):
            logit = self.logit [idx][:len (label)]
            y = self.y [idx][:len (label)]
            probs = predutil.softmax (logit)
            prob = "{:.5f}".format (probs [np.argmax (probs)])
            if prob not in stat:
                stat [prob] = [0, 0]
            if np.argmax (probs) == np.argmax (y):    
                stat [prob][0] += 1
            stat [prob][1] += 1
            
        ordered = [] 
        accum = 0        
        accum_t = 0
        total = len (self.y)
        for lev, (c, t) in sorted (stat.items (), reverse = True):            
            accum += c            
            accum_t += t
            accuracy = accum / accum_t * 100
            ordered.append ((
                lev, 
                accum, accum_t, accuracy, 
                accum / total * 100 
            ))
            if len (ordered) >= 10 and accuracy < 100.:
                break
        return ordered    
    



#
# Copyright (c) Microsoft. All rights reserved.
# Licensed under the MIT license. See LICENSE.md file in the project root for full license information.
#

import numpy as np
import random as rnd
from PIL import Image
from scipy import ndimage
from .rect_util import Rect
from .transform import contrast_stretching
import cv2

def compute_norm_mat(base_width, base_height): 
    # normalization matrix used in image pre-processing 
    x      = np.arange(base_width)
    y      = np.arange(base_height)
    X, Y   = np.meshgrid(x, y)
    X      = X.flatten()
    Y      = Y.flatten() 
    A      = np.array([X*0+1, X, Y]).T 
    A_pinv = np.linalg.pinv(A)
    return A, A_pinv

def preproc_img(img, A, A_pinv):
    # compute image histogram 
    img_flat = img.flatten()
    img_hist = np.bincount(img_flat, minlength = 256)

    # cumulative distribution function 
    cdf = img_hist.cumsum() 
    cdf = cdf * (2.0 / cdf[-1]) - 1.0 # normalize 

    # histogram equalization 
    img_eq = cdf[img_flat] 

    diff = img_eq - np.dot(A, np.dot(A_pinv, img_eq))

    # after plane fitting, the mean of diff is already 0 
    std = np.sqrt(np.dot(diff,diff)/diff.size)
    if std > 1e-6: 
        diff = diff/std
    return diff.reshape(img.shape)

NORM_MATRIX = {}
def normalize (img_arr, reshape = None):
    global NORM_MATRIX

    w, h = img_arr.shape [:2]
    if (w, h) not in NORM_MATRIX:
        NORM_MATRIX [(w, h)] = compute_norm_mat (w, h)
    A, A_pinv = NORM_MATRIX [(w, h)]
    final_image = preproc_img (img_arr, A = A, A_pinv = A_pinv)
    if reshape:
       return final_image.reshape (reshape)
    return final_image   

def distort_img(img, roi = None, out_width = None, out_height = None, max_shift = 0.0, max_scale = 1.0, max_angle = 0.0, max_skew = 0.0, flip = False, noise = False, random_contrast = False):
    default_width, default_height = img.width, img.height
    
    roi = roi or Rect([0, 0, default_width, default_height])
    out_width = out_width or default_width
    out_height = out_height or default_height

    shift_y = out_height*max_shift*rnd.uniform(-1.0,1.0)
    shift_x = out_width*max_shift*rnd.uniform(-1.0,1.0)

    # rotation angle 
    angle = max_angle*rnd.uniform(-1.0,1.0)

    #skew 
    sk_y = max_skew*rnd.uniform(-1.0, 1.0)
    sk_x = max_skew*rnd.uniform(-1.0, 1.0)

    # scale 
    scale_y = rnd.uniform(1.0, max_scale) 
    if rnd.choice([True, False]): 
        scale_y = 1.0/scale_y 
    scale_x = rnd.uniform(1.0, max_scale) 

    if rnd.choice([True, False]): 
        scale_x = 1.0/scale_x 
    T_im = crop_img(img, roi, out_width, out_height, shift_x, shift_y, scale_x, scale_y, angle, sk_x, sk_y)
    if flip and rnd.choice([True, False]): 
        T_im = np.fliplr(T_im)
    if noise and rnd.choice([True, False]):
        T_im = add_noise (T_im, rnd.choice (['gauss', 's&p', 'poisson', 'speckle']))
        T_im = np.clip (T_im.astype ('int64'), 0, 255)                
    if random_contrast and rnd.choice([True, False]):
        T_im = contrast_stretching (T_im, rnd.randrange (20), rnd.randrange (80, 101))
    return T_im

def crop_img(img, roi, crop_width, crop_height, shift_x, shift_y, scale_x, scale_y, angle, skew_x, skew_y):
    # current face center 
    ctr_in = np.array((roi.center().y, roi.center().x))
    ctr_out = np.array((crop_height/2.0+shift_y, crop_width/2.0+shift_x))
    out_shape = (crop_height, crop_width)
    s_y = scale_y*(roi.height()-1)*1.0/(crop_height-1)
    s_x = scale_x*(roi.width()-1)*1.0/(crop_width-1)
    
    # rotation and scale 
    ang = angle*np.pi/180.0 
    transform = np.array([[np.cos(ang), -np.sin(ang)], [np.sin(ang), np.cos(ang)]])
    transform = transform.dot(np.array([[1.0, skew_y], [0.0, 1.0]]))
    transform = transform.dot(np.array([[1.0, 0.0], [skew_x, 1.0]]))
    transform = transform.dot(np.diag([s_y, s_x]))
    offset = ctr_in-ctr_out.dot(transform)
    
    # each point p in the output image is transformed to pT+s, where T is the matrix and s is the offset
    T_im = ndimage.interpolation.affine_transform(input = img, 
                                                  matrix = np.transpose(transform), 
                                                  offset = offset, 
                                                  output_shape = out_shape, 
                                                  order = 1,   # bilinear interpolation 
                                                  mode = 'reflect', 
                                                  prefilter = False)
    return T_im

def add_noise (image, noise_typ = 'gauss'):    
   if noise_typ == "gauss":
      mean = 0
      var = 0.1
      sigma = var**0.5
      gauss = np.random.normal(mean,sigma,image.shape)      
      gauss = gauss.reshape(*image.shape)
      noisy = image + gauss
      return noisy

   elif noise_typ == "s&p":
      s_vs_p = 0.5
      amount = 0.004
      out = np.copy(image)
      # Salt mode
      num_salt = np.ceil(amount * image.size * s_vs_p)
      coords = [np.random.randint(0, i - 1, int(num_salt))
              for i in image.shape]
      out[coords] = 1

      # Pepper mode
      num_pepper = np.ceil(amount* image.size * (1. - s_vs_p))
      coords = [np.random.randint(0, i - 1, int(num_pepper))
              for i in image.shape]
      out[coords] = 0
      return out

   elif noise_typ == "poisson":
      vals = len(np.unique(image))
      vals = 2 ** np.ceil(np.log2(vals))
      noisy = np.random.poisson(image * vals) / float(vals)
      return noisy

   elif noise_typ =="speckle":      
      gauss = np.random.randn(*image.shape)
      gauss = gauss.reshape(*image.shape)        
      noisy = image + image * gauss
      return noisy



import cv2
import numpy as np

def average_hash (arr, size = 16):
    geryscale = cv2.resize (arr, (size, size))
    if geryscale.shape [-1] > 1:
        geryscale = cv2.cvtColor(geryscale, cv2.COLOR_BGR2GRAY)
    return 1 * (geryscale > geryscale.mean ())
    
def hamming_dist (a, b):
    return (a != b).sum () / np.product (a.shape)


        


# Default libraries
import os

try:
    import numpy as np
    import librosa    # Library for finding silence interval.
except ImportError:
    print("Please install numpy and librosa packages.")
    
def read(filename):
    # Read file and then return wave sequence and its frequency.
    try:
        wave, fq = librosa.load(filename)
    except FileNotFoundError:
        print("File {} is not found.".format(filename))
        wave, fq = None, None
    return (wave, fq)

def detect(wave, threshold=20):
    # Find the non_silence windows.
    # Return value is a list of Boolean data.
    # If the energy of ith interval of wave file is less then threshold then
    # it returns False at the ith component and vice versa. 
    return librosa.effects._signal_to_frame_nonsilent(wave, top_db=threshold)

def false_count(detection):
    # Input is the list of Boolean data.
    # It returns starting window of silent interval and how long it lasts.
    # For example if in put is [T T T F F F F T T F F T T F], then it returns
    # [[3,4],[9,2],[13,1]]
    i = 0
    detect_list = []
    while i < len(detection):
        if detection[i] == True:
            i += 1
        else:
            j = 0
            while (i + j) < len(detection) and detection[i + j] == False:
                j += 1
            detect_list.append((i, j))
            i += j
    return detect_list

def silence_index(wave, threshold=20, min_silence_frames=50):
    # Input is a wave file and threshold and nim_silence_frames.
    # Using false_count, we get the the silent information of wave file.
    # It returns the silent interval larger than min_silence_frames.
    index = []
    detect_list = false_count(detect(wave, threshold))
    for (i, j) in detect_list:
        if j >= min_silence_frames:
            index.append((i * 512, (i+j-1)*512+2048-1))
    return index

def clip_cut(wave, threshold=20, min_silence_frames=50):
    # Using the result of silence_index, we can find the non_silence interval.
    index = []
    wave_len = len(wave)//512*512
    silence_list=silence_index(wave, threshold, min_silence_frames)
    if not silence_list:
        return index 
    if silence_list[0][0] > 0:
        index.append((0,silence_list[0][0]-1))
    for k in range(len(silence_list)-1):
        index.append((silence_list[k][1]+1,silence_list[k+1][0]-1))
    if silence_list[-1][1] < wave_len-1:
        index.append((silence_list[-1][1]+1, wave_len-1))
    return index

def frame_to_time(frame_stamp, frequency, wave_len, time_laps=0.2):
    # It returns lists of time_interval from frame_interval.
    # We will add time_laps at each starting and end point of time_interval.
    stamp = []
    for k in range(len(frame_stamp)):
        if (k == 0) and (frame_stamp[k][0]/frequency < time_laps):
            stamp.append((0, frame_stamp[k][1]/frequency+time_laps))
        elif (k == len(frame_stamp) - 1) and (frame_stamp[k][0]/frequency + time_laps > wave_len/frequency):
            stamp.append((frame_stamp[k][0]/frequency-time_laps, wave_len/frequency))
        else:
            stamp.append((frame_stamp[k][0]/frequency-time_laps, frame_stamp[k][1]/frequency+time_laps))
    return stamp



import cv2
from mtcnn.mtcnn import MTCNN
import os
from . import measure
from ..video import vtool

DETECTOR = None

def create_detector ():
    global DETECTOR    
    if DETECTOR is None:
        DETECTOR = MTCNN ()
    
def load (pixels, name):
    create_detector ()
    if pixels.shape [-1] != 3:
        assert name, "Parameter name is required"
        temp = ".{}-temp.bmp".format (name)
        try:
            cv2.imwrite (temp, pixels)
            pixels = cv2.imread (temp)
        finally:    
            os.remove (temp)
    return pixels
    
def detect (pixels, name = 'face-detector', resize = (48, 48)):
    # crop and strech
    image = load (pixels, name)    
    result = DETECTOR.detect_faces (image)
    img = None
    if result:
        bb = result [0]['box']
        x = max (0, bb [0])
        y = max (0, bb [1])        
        crop_img = image [y:y + bb [3], x:x + bb [2]]
        if resize [0] == -1:
            resize = (int (bb [2] / (bb [3] / resize [1])), resize [1])
        elif resize [1] == -1:
            resize = (resize [0], int (bb [3] / (bb [2] / resize [0])))
        img = cv2.resize (crop_img, resize)
    return result, img

def detect_keep_ratio (pixels, name = 'face-detector', resize = (48, 48)):
    image = load (pixels, name)
    result = DETECTOR.detect_faces (image)
    img = None    
    if result:
        iw, ih = image.shape [:2]

        bb = result [0]['box']
        x = max (0, bb [0])
        y = max (0, bb [1])
        x_ = min (iw, x + bb [2])
        y_ = min (ih, y + bb [3])        
        width = x_ - x
        height = y_ - y
        
        # pad 15%
        y = max (0, y - int (height * 0.06))
        y_ = min (ih, y_ + int (height * 0.06))
        height = y_ - y

        if width > height:
            pad = (width - height) // 2
            y -= pad; y_ += pad
            y = max (0, y); y_ = min (ih, y_)
        else:
            pad = (height - width) // 2
            x -= pad; x_ += pad
            x = max (0, x); x_ = min (iw, x_)            
        
        print (bb, x, y, x_, y_)
        crop_img = image [y:y_, x:x_]          
        img = cv2.resize (crop_img, resize)
    return result, img

def mark (pixels, name = None):
    image = load (pixels, name)    
    result = DETECTOR.detect_faces (image)
    if not result:
        return
    else:
        bounding_box = result [0]['box']
        keypoints = result[0]['keypoints']
        cv2.rectangle (image,
                      (bounding_box[0], bounding_box[1]),
                      (bounding_box[0]+bounding_box[2], bounding_box[1] + bounding_box[3]),
                      (0,155,255),
                      2)
        cv2.circle (image, (keypoints['left_eye']), 2, (0,155,255), 2)
        cv2.circle (image, (keypoints['right_eye']), 2, (0,155,255), 2)
        cv2.circle (image, (keypoints['nose']), 2, (0,155,255), 2)
        cv2.circle (image, (keypoints['mouth_left']), 2, (0,155,255), 2)
        cv2.circle (image, (keypoints['mouth_right']), 2, (0,155,255), 2)    
    return image


def from_video_with_metric (video, frame_skip = 10, min_width = 48, min_dist = 0.1, resize = (48, 48), with_marked = False, choice = 'maxdistance', orientation = 0):
    hashes = []
    faces = []
    hashmap = {}
    markeds = []
    cluster = []
    for idx, pixels in enumerate (vtool.capture (video, frame_skip)):
        if orientation:
            (h, w) = pixels.shape[:2]     
            center = (w / 2, h / 2) 
            M = cv2.getRotationMatrix2D (center, orientation, 1.0)
            pixels = cv2.warpAffine (pixels, M, (h, w))  

        result, img = detect (pixels, resize = resize)
        if not result:
            continue
        bb = result [0]['box']
        if bb [2] < min_width:        
            continue        
        h = measure.average_hash (img)
        hashmap [id (h)] = (img, pixels)
        
        if choice == "maxdistance":
            dup = False
            for h_ in hashes:
                dist = measure.hamming_dist (h, h_)
                if dist < min_dist:
                    #print ("threshold", dist)
                    dup = True
                    break
    
            hashes.append (h)
            if not dup:
                faces.append ((result, img))
                if with_marked:
                    markeds.append (mark (pixels))
        
        elif choice == "all":
            faces.append ((result, img))
            if with_marked:
                markeds.append (mark (pixels))
            
        elif choice == "single":
            if len (cluster) == 0:
                cluster.append ([h])
                continue
            
            clustered = False
            for hashes in cluster:                
                for h_ in hashes:
                    dist = measure.hamming_dist (h, h_)
                    if dist < min_dist:
                        hashes.append (h)
                        clustered = True                        
                        break
            
            if not clustered: # new cluster
                cluster.append ([h])
        
    if choice == "single" and cluster:        
        for h in sorted (cluster, key = lambda x: len (x))[-1]:
            img, pixels = hashmap [id (h)]
            faces.append ((result, img))            
            if with_marked:
                markeds.append (mark (pixels))
                    
    if with_marked:
        return faces, markeds
    
    return faces

def from_video (video, frame_skip = 10, min_width = 48, min_dist = 0.1, resize = (48, 48), with_marked = False, choice = 'maxdistance', orientation = 0):
    faces = from_video_with_metric (video, frame_skip, min_width, min_dist, resize, with_marked, choice, orientation)
    return [img for metric, img in faces]
    



import sys
import numpy as np

class Overfit:
    def __init__ (self, keep_count = 0, period = 5, min_validations = 0):      
        self.keep_count = keep_count
        self.period = period
        self.min_validations = min_validations
        
        self.validations = 0
        self.overfitted_count = 0
        self.cost_log = [[], []]        
        self.new_record = False
        self.min_cost = sys.maxsize
        self.overfit = False
    
    @property
    def latest_cost (self):
        return  self.cost_log [1][-1]
        
    def is_overfit (self):
        return self.overfit
    
    def is_renewaled (self):
        return self.new_record
    
    def add_cost (self, cost, is_validating):
        if not is_validating: # currently recall loss is not used
            return
        
        self.validations += 1       
        self.overfit, self.new_record = False, False        
        if is_validating:
            if cost < self.min_cost: 
                self.min_cost = cost   
                self.new_record = True
                self.overfitted_count = 0 # reset count
        
        index = int (is_validating)
        if self.cost_log [index]:
            latest = np.mean (self.cost_log [index])
        self.cost_log [index].append (cost)
        if len (self.cost_log  [index]) < self.period:
            return
        
        self.cost_log [index] = self.cost_log [index] [-self.period:]        
        if self.min_validations and self.validations < self.min_validations:
            return
        
        current = np.mean (self.cost_log [index])
        if current >= latest:
            self.overfitted_count += 1            
            if self.keep_count and self.overfitted_count > self.keep_count:
                self.overfit = True        
        else:
            self.overfitted_count = 0
        
        #print (self.cost_log)
        #print (latest, current)
        #print (self.overfitted_count, self.overfit)
        


"""
Most codes from https://github.com/musyoku/adversarial-autoencoder/blob/master/aae/sampler.py
"""
import numpy as np
from math import sin,cos,sqrt


def uniform(batch_size, n_dim, n_labels=10, minv=-1, maxv=1, label_indices=None):
    if label_indices is not None:
        if n_dim != 2 or n_labels != 10:
            raise Exception("n_dim must be 2 and n_labels must be 10.")

        def sample(label, n_labels):
            num = int(np.ceil(np.sqrt(n_labels)))
            size = (maxv-minv)*1.0/num
            x, y = np.random.uniform(-size/2, size/2, (2,))
            i = label / num
            j = label % num
            x += j*size+minv+0.5*size
            y += i*size+minv+0.5*size
            return np.array([x, y]).reshape((2,))

        z = np.empty((batch_size, n_dim), dtype=np.float32)
        for batch in range(batch_size):
            for zi in range((int)(n_dim/2)):
                    z[batch, zi*2:zi*2+2] = sample(label_indices[batch], n_labels)
    else:
        z = np.random.uniform(minv, maxv, (batch_size, n_dim)).astype(np.float32)
    return z

def gaussian(batch_size, n_dim, mean=0, var=1, n_labels=10, use_label_info=False):
    if use_label_info:
        if n_dim != 2 or n_labels != 10:
            raise Exception("n_dim must be 2 and n_labels must be 10.")

        def sample(n_labels):
            x, y = np.random.normal(mean, var, (2,))
            angle = np.angle((x-mean) + 1j*(y-mean), deg=True)
            dist = np.sqrt((x-mean)**2+(y-mean)**2)

            # label 0
            if dist <1.0:
                label = 0
            else:
                label = ((int)((n_labels-1)*angle))//360

                if label<0:
                    label+=n_labels-1

                label += 1

            return np.array([x, y]).reshape((2,)), label

        z = np.empty((batch_size, n_dim), dtype=np.float32)
        z_id = np.empty((batch_size), dtype=np.int32)
        for batch in range(batch_size):
            for zi in range((int)(n_dim/2)):
                    a_sample, a_label = sample(n_labels)
                    z[batch, zi*2:zi*2+2] = a_sample
                    z_id[batch] = a_label
        return z, z_id
    else:
        z = np.random.normal(mean, var, (batch_size, n_dim)).astype(np.float32)
        return z

def gaussian_mixture(batch_size, n_dim=2, n_labels=10, x_var=0.5, y_var=0.1, label_indices=None):
    if n_dim != 2:
        raise Exception("n_dim must be 2.")

    def sample(x, y, label, n_labels):
        shift = 1.4
        r = 2.0 * np.pi / float(n_labels) * float(label)
        new_x = x * cos(r) - y * sin(r)
        new_y = x * sin(r) + y * cos(r)
        new_x += shift * cos(r)
        new_y += shift * sin(r)
        return np.array([new_x, new_y]).reshape((2,))

    x = np.random.normal(0, x_var, (batch_size, (int)(n_dim/2)))
    y = np.random.normal(0, y_var, (batch_size, (int)(n_dim/2)))
    z = np.empty((batch_size, n_dim), dtype=np.float32)
    for batch in range(batch_size):
        for zi in range((int)(n_dim/2)):
            if label_indices is not None:
                z[batch, zi*2:zi*2+2] = sample(x[batch, zi], y[batch, zi], label_indices[batch], n_labels)
            else:
                z[batch, zi*2:zi*2+2] = sample(x[batch, zi], y[batch, zi], np.random.randint(0, n_labels), n_labels)

    return z

def swiss_roll(batch_size, n_dim=2, n_labels=10, label_indices=None):
    if n_dim != 2:
        raise Exception("n_dim must be 2.")

    def sample(label, n_labels):
        uni = np.random.uniform(0.0, 1.0) / float(n_labels) + float(label) / float(n_labels)
        r = sqrt(uni) * 3.0
        rad = np.pi * 4.0 * sqrt(uni)
        x = r * cos(rad)
        y = r * sin(rad)
        return np.array([x, y]).reshape((2,))

    z = np.zeros((batch_size, n_dim), dtype=np.float32)
    for batch in range(batch_size):
        for zi in range((int)(n_dim/2)):
            if label_indices is not None:
                z[batch, zi*2:zi*2+2] = sample(label_indices[batch], n_labels)
            else:
                z[batch, zi*2:zi*2+2] = sample(np.random.randint(0, n_labels), n_labels)
    return z


"""
Hans Roh 2015 -- http://osp.skitai.com
License: BSD
"""
import re
import sys
import os
import shutil, glob
from warnings import warn
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

with open('dnn/__init__.py', 'r') as fd:
	version = re.search(r'^__version__\s*=\s*"(.*?)"',fd.read(), re.M).group(1)
	
if sys.argv[-1] == 'publish':
	buildopt = ['sdist', 'upload']
	if os.name == "nt":
		buildopt.insert (0, 'bdist_wheel')
	os.system('python setup.py %s' % " ".join (buildopt))
	#os.system('twine upload dist/aquests-%s*' % version)
	for each in os.listdir ("dist"):
		os.remove (os.path.join ('dist', each))
	sys.exit()

classifiers = [
  'License :: OSI Approved :: MIT License',
  'Development Status :: 3 - Alpha',  
	'Environment :: Console',	
	'Topic :: Software Development :: Libraries :: Python Modules',
	'Topic :: Scientific/Engineering :: Artificial Intelligence',
	'Intended Audience :: Developers',
	'Intended Audience :: Science/Research',
	'Programming Language :: Python',	
	'Programming Language :: Python :: 3'
]

packages = [
	'dnn',	
	'dnn.audio',
	'dnn.video',
	'dnn.image',
	'dnn.text',
	'dnn.similarity',
	'dnn.costs',
]

package_dir = {'dnn': 'dnn'}

package_data = {
	"dnn": []
}

install_requires = [	
	"rs4",
	"matplotlib",
	"sklearn",
	"scikit-image",
	"imageio",	
	"future", 
	"networkx",
	"tqdm",
	"librosa", 
	"soundfile", 	
	"colorama",
        "hyperopt"
]

with open ('README.rst', encoding='utf-8') as f:
	long_description = f.read ()
	
setup(
	name='dnn',
	version=version,
	description='Deep Neural Network Library',
	long_description = long_description,
	url = 'https://gitlab.com/hansroh/dnn',
	author='Hans Roh',	
	author_email='hansroh@gmail.com',	
	packages=packages,
	package_dir=package_dir,
	package_data = package_data,
	license='MIT',
	platforms = ["posix",],
	download_url = "https://pypi.python.org/pypi/dnn",
	install_requires = install_requires,
	classifiers=classifiers
)
